import atexit
from dataclasses import dataclass, fields
import sys
import os.path
from datetime import datetime, timedelta
import contextlib
import asyncio
from typing import OrderedDict, Any, List, Callable, Optional
from scabha.basetypes import EmptyListDefault
from omegaconf import OmegaConf
import psutil

import rich.progress
import rich.logging
from rich.table import Table
from rich.text import Text

from stimela import stimelogging

# this is "" for the main process, ".0", ".1", for subprocesses, ".0.0" for nested subprocesses
_subprocess_identifier = ""

def get_subprocess_id():
    return _subprocess_identifier

def add_subprocess_id(num: int):
    global _subprocess_identifier
    _subprocess_identifier += f".{num}"

progress_bar = progress_task = None

_start_time = datetime.now()
_prev_disk_io = None, None

@dataclass
class TaskInformation(object):
    names: List[str]
    status: str = ""
    task_attrs: List[str] = EmptyListDefault()
    command: Optional[str] = None
    status_reporter: Optional[Callable] = None
    hide_local_metrics: bool = False

    def __post_init__(self):
        self.names_orig = list(self.names)

    @property
    def description(self):
        name = '.'.join(self.names)
        # if self.status:
        #     name += f": [dim]{self.status}[/dim]"
        ## OMS: omit attributes from task status for now
        # if self.task_attrs:
        #     name += f"\[{', '.join(self.task_attrs)}]"
        return name

# stack of task information -- most recent subtask is at the end
_task_stack = []

def init_progress_bar(boring=False):
    global progress_console, progress_bar, progress_task
    progress_console = rich.console.Console(file=sys.stdout, highlight=False, emoji=False)
    progress_bar = rich.progress.Progress(
        rich.progress.SpinnerColumn(),
        "[yellow]{task.fields[elapsed_time]}[/yellow]",
        "[bold]{task.description}[/bold]",
        rich.progress.SpinnerColumn(),
        "[dim]{task.fields[status]}[/dim]",
        "{task.fields[command]}",
        rich.progress.TimeElapsedColumn(),
        "{task.fields[cpu_info]}",
        refresh_per_second=2,
        console=progress_console,
        transient=True,
        disable=boring)

    progress_task = progress_bar.add_task("stimela", status="", command="starting", cpu_info=" ", elapsed_time="", start=True)
    progress_bar.__enter__()
    atexit.register(destroy_progress_bar)
    return progress_bar, progress_console

def destroy_progress_bar():
    global progress_bar
    if progress_bar is not None:
        progress_bar.__exit__(None, None, None)
        progress_bar = None

def restate_progress():
    """Renders a snapshot of the progress bar onto the console"""
    if progress_bar is not None:
        progress_console.print(progress_bar.get_renderable())
        progress_console.rule()


@contextlib.contextmanager
def declare_subtask(subtask_name, status_reporter=None, hide_local_metrics=False):
    task_names = []
    if _task_stack:
        task_names = _task_stack[-1].names + \
                    (_task_stack[-1].task_attrs or [])
    task_names.append(subtask_name)
    _task_stack.append(
        TaskInformation(task_names, status_reporter=status_reporter, hide_local_metrics=hide_local_metrics)
    )
    update_process_status()
    try:
        yield subtask_name
    finally:
        _task_stack.pop(-1)
        update_process_status()


def declare_subtask_status(status):
    _task_stack[-1].status = status
    update_process_status()


def declare_subtask_attributes(*args, **kw):
    _task_stack[-1].task_attrs = [str(x) for x in args] + \
                                 [f"{key} {value}" for key, value in kw.items()]
    update_process_status()


class _CommandContext(object):
    def __init__(self, command):
        self.command = command
        _task_stack[-1].command = command
        update_process_status()
    def ctrl_c(self):
        _task_stack[-1].command = f"{self.command}(^C)"
        update_process_status()
    def update_status(self, status):
        _task_stack[-1].command = f"{self.command} ({status})"
        update_process_status()


@contextlib.contextmanager
def declare_subcommand(command):
    progress_bar and progress_bar.reset(progress_task)
    try:
        yield _CommandContext(command)
    finally:
        _task_stack[-1].command = None
        update_process_status()


@dataclass
class TaskStatsDatum(object):
    cpu: float          = 0
    mem_used: float     = 0
    mem_total: float    = 0
    load: float         = 0
    read_count: int     = 0
    read_gb: float      = 0
    read_gbps: float    = 0
    read_ms: float      = 0
    write_count: int    = 0
    write_gb: float     = 0 
    write_gbps: float   = 0 
    write_ms: float     = 0
    num_samples: int    = 0

    def __post_init__(self):
        self.extras = []

    def insert_extra_stats(self, **kw):
        for name, value in kw.items():
            self.extras.append(name)
            setattr(self, name, value)

    def add(self, other: "TaskStatsDatum"):
        for f in _taskstats_sample_names + other.extras:
            if not hasattr(self, f):
                self.extras.append(f)
            setattr(self, f, getattr(self, f, 0) + getattr(other, f))

    def peak(self, other: "TaskStatsDatum"):
        for f in _taskstats_sample_names + other.extras:
            if hasattr(self, f):
                setattr(self, f, max(getattr(self, f, -1e-9999), getattr(other, f)))
            else:
                self.extras.append(f)
                setattr(self, f, getattr(other, f))

    def averaged(self):
        avg = TaskStatsDatum(num_samples=1)
        for f in _taskstats_sample_names:
            setattr(avg, f, getattr(self, f) / self.num_samples)
        avg.insert_extra_stats(**{name: getattr(self, name) / self.num_samples for name in self.extras})
        return avg


_taskstats_sample_names = [f.name for f in fields(TaskStatsDatum)]

_taskstats = OrderedDict()
_task_start_time = OrderedDict()


def collect_stats():
    """Returns dictionary of per-task stats (elapsed time, sums, peaks)"""
    # cumulative add -- substeps contribute to parent steps
    for key in list(_taskstats.keys())[::-1]:
        _, sum, peak = _taskstats[key]
        key1 = tuple(key[:-1])
        if key1 in _taskstats:
            _, sum1, peak1 = _taskstats[key1]
            sum1.add(sum)
            peak1.peak(peak)
    return _taskstats


def add_missing_stats(stats):
    """Adds stats that weren't recorded into dictionary"""
    for key, value in stats.items():
        if key not in _taskstats:
            _taskstats[key] = value


def stats_field_names():
    return _taskstats_sample_names


def update_stats(now: datetime, sample: TaskStatsDatum):
    if _task_stack:
        ti = _task_stack[-1]
        keys = [tuple(ti.names)]
        if ti.task_attrs:
            keys.append(tuple(ti.names + ti.task_attrs))
    else:
        keys = [()]

    for key in keys:
        _, sum, peak = _taskstats.setdefault(key, [0, TaskStatsDatum(), TaskStatsDatum()])
        sum.add(sample)
        peak.peak(sample)
        start = _task_start_time.setdefault(key, now)
        _taskstats[key][0] = (now - start).total_seconds()


def update_process_status():
    # current subtask info
    ti = _task_stack[-1] if _task_stack else None

    # elapsed time since start
    now = datetime.now()
    elapsed = str(now - _start_time).split('.', 1)[0]

    # form up sample datum
    s = TaskStatsDatum(num_samples=1)
    # CPU and memory
    s.cpu = psutil.cpu_percent()
    mem = psutil.virtual_memory()
    s.mem_used = round(mem.total*mem.percent/100 / 2**30)
    s.mem_total = round(mem.total / 2**30)
    # load
    s.load, _, _ = psutil.getloadavg()

    # get disk I/O stats
    disk_io = psutil.disk_io_counters()
    global _prev_disk_io
    prev_io, prev_time = _prev_disk_io 
    if prev_io is not None:
        delta = (now - prev_time).total_seconds()
        io = {}
        for key in 'read_bytes', 'read_count', 'read_time', 'write_bytes', 'write_count', 'write_time':
            io[key] = getattr(disk_io, key) - getattr(prev_io, key)
        s.read_count = io['read_count']
        s.write_count = io['write_count']
        s.read_gb = io['read_bytes']/2**30
        s.write_gb = io['write_bytes']/2**30
        s.read_gbps = s.read_gb / delta
        s.write_gbps = s.write_gb / delta
        s.read_ms = io['read_time'] 
        s.write_ms = io['write_time']
    else:
        io = None
    _prev_disk_io = disk_io, now

    # call extra status reporter
    if ti and ti.status_reporter:
        extra_metrics, extra_stats = ti.status_reporter()
        if extra_stats:
            s.insert_extra_stats(**extra_stats)
    else:
        extra_metrics = None

    # if a progress bar exists, update it
    if progress_bar is not None:
        cpu_info = []
        # add local metering, if not diabled by a task in the stack
        if not any(t.hide_local_metrics for t in _task_stack):
            cpu_info = [
                f"CPU [green]{s.cpu:2.1f}%[/green]",
                f"RAM [green]{round(s.mem_used):3}[/green]/[green]{round(s.mem_total)}[/green]G",
                f"Load [green]{s.load:2.1f}[/green]" 
            ]

            if io is not None:
                cpu_info += [
                    f"R [green]{s.read_count:-4}[/green] [green]{s.read_gbps:2.2f}[/green]G [green]{s.read_ms:4}[/green]ms",
                    f"W [green]{s.write_count:-4}[/green] [green]{s.write_gbps:2.2f}[/green]G [green]{s.write_ms:4}[/green]ms "
                ]
        # add extra metering
        cpu_info += extra_metrics or []

        updates = dict(elapsed_time=elapsed, cpu_info="|".join(cpu_info))

        if ti is not None:
            updates['description'] = ti.description
            updates['status'] = ti.status or ''
            updates['command'] = ti.command or ''

        progress_bar.update(progress_task, **updates)

    # update stats
    update_stats(now, s)


async def run_process_status_update():
    if progress_bar:
        with contextlib.suppress(asyncio.CancelledError):
            while True:
                update_process_status()
                await asyncio.sleep(1)

_printed_stats = dict(
    k8s_cores="k8s cores",
    k8s_mem="k8s mem GB",
    cpu="CPU %",
    mem_used="Mem GB",
    load="Load",
    read_gbps="R GB/s",
    write_gbps="W GB/s",
    )

# these stats are written as sums
_sum_stats = ("read_count", "read_gb", "read_ms", "write_count", "write_gb", "write_ms")

def render_profiling_summary(stats: TaskStatsDatum, max_depth, unroll_loops=False):

    table_avg = Table(title=Text("\naverages & total I/O", style="bold"))
    table_avg.add_column("")
    table_avg.add_column("time hms", justify="right")

    table_peak = Table(title=Text("\npeaks", style="bold"))
    table_peak.add_column("")
    table_peak.add_column("time hms", justify="right")
    
    # accumulate set of all stats available
    available_stats = set(_taskstats_sample_names)
    for name, (elapsed, sum, peak) in stats.items():
        available_stats.update(sum.extras)
        available_stats.update(peak.extras)

    for f, label in _printed_stats.items():
        if f in available_stats:
            table_avg.add_column(label, justify="right")
            table_peak.add_column(label, justify="right")

    table_avg.add_column("R GB", justify="right")
    table_avg.add_column("W GB", justify="right")

    for name_tuple, (elapsed, sum, peak) in stats.items():
        if name_tuple and len(name_tuple) <= max_depth:
            # skip loop iterations, if not unrolling loops 
            if not unroll_loops and any(n.endswith("]") for n in name_tuple):
                continue
            secs, mins, hours = elapsed % 60, int(elapsed // 60) % 60, int(elapsed // 3600)
            tstr = f"{hours:d}:{mins:02d}:{secs:04.1f}"
            avg = sum.averaged()
            indentation_level = len(name_tuple) - 1
            avg_row = ["  " * indentation_level + name_tuple[-1], tstr]
            peak_row = avg_row.copy()
            for f, label in _printed_stats.items():
                if f in available_stats:
                    avg_row.append(f"{getattr(avg, f):.2f}" if hasattr(avg, f) else "")
                    peak_row.append(f"{getattr(peak, f):.2f}" if hasattr(peak, f) else "")

            avg_row += [f"{sum.read_gb:.2f}", f"{sum.write_gb:.2f}"]
            table_avg.add_row(*avg_row)
            table_peak.add_row(*peak_row)

    stimelogging.declare_chapter("profiling results")
    destroy_progress_bar()
    from rich.columns import Columns
    # progress_console.print(table_avg, justify="center") 
    # progress_console.print(table_peak, justify="center")  
    # progress_console.print(Columns((table_avg, table_peak)), justify="center") 

    with progress_console.capture() as capture:
        progress_console.print(Columns((table_avg, table_peak)), justify="center") 

    text = capture.get()

    return text

    
# from rich.console import Console
# console = Console()
# with console.capture() as capture:
#     console.print("[bold red]Hello[/] World")
# str_output = capture.get()

def save_profiling_stats(log, print_depth=2, unroll_loops=False):
    from . import stimelogging
    
    stats = collect_stats()
    summary = render_profiling_summary(stats, print_depth, unroll_loops=unroll_loops)
    if print_depth:
        print(summary)

    filename = os.path.join(stimelogging.get_logfile_dir(log) or '.', "stimela.stats.full")

    stats_dict = OmegaConf.create()

    for name, (elapsed, sum, peak) in stats.items():
        if name:
            name = '.'.join(name)
            avg = sum.averaged()
            davg = {f: getattr(avg, f) for f in _taskstats_sample_names}
            dpeak = {f: getattr(peak, f) for f in _taskstats_sample_names}
            dsum = {f: getattr(sum, f) for f in _sum_stats}

            stats_dict[name] = dict(elapsed=elapsed, avg=davg, peak=dpeak, total=dsum)

    OmegaConf.save(stats_dict, filename)

    log.info(f"saved full profiling stats to {filename}")

    filename = os.path.join(stimelogging.get_logfile_dir(log) or '.', "stimela.stats.summary.txt")
    open(filename, "wt").write(summary)

    log.info(f"saved summary to {filename}")



import logging, datetime, resource, os.path

from typing import Dict, Optional, Any

import stimela
import stimela.kitchen
from stimela.utils.xrun_asyncio import xrun
from stimela.exceptions import StimelaProcessRuntimeError, BackendSpecificationError
from scabha.substitutions import substitutions_from


def update_rlimits(rlimits: Dict[str, Any], log: logging.Logger):
    for name, limit in rlimits.items():
        rname = f"RLIMIT_{name}"
        if not hasattr(resource, rname):
            raise StimelaProcessRuntimeError(f"unknown resource limit 'backend.rlimits.{name}'")
        rconst = getattr(resource, rname)
        # get current limits
        soft, hard = resource.getrlimit(rconst)
        # check for unlimited
        if limit is None:
            limit = resource.RLIM_INFINITY
            if hard != resource.RLIM_INFINITY:
                raise StimelaProcessRuntimeError(f"can't set backend.rlimits.{name}=unlimited: hard limit is {hard}")
        else:
            if limit > hard:
                raise StimelaProcessRuntimeError(f"can't set backend.rlimits.{name}={limit}: hard limit is {hard}")
        resource.setrlimit(rconst, (limit, hard))
        log.debug(f"setting soft limit {name}={limit} (hard limit is {hard})")



def build_command_line(cab: 'stimela.kitchen.cab.Cab', params: Dict[str, Any], subst: Optional[Dict[str, Any]] = None,
                        virtual_env: Optional[str] = None, log: Optional[logging.Logger] = None):
    return cab.flavour.get_arguments(cab, params, subst, virtual_env=virtual_env, log=log)


def run(cab: 'stimela.kitchen.cab.Cab', params: Dict[str, Any], fqname: str,
        backend: 'stimela.backend.StimelaBackendOptions',
        log: logging.Logger, subst: Optional[Dict[str, Any]] = None,
        wrapper: Optional['stimela.backends.runner.BackendWrapper'] = None):
    """
    Runs cab contents

    Args:
        cab: cab object
        params: cab parameters
        backend: backed settings object
        log (logger): logger to use
        subst (Optional[Dict[str, Any]]): Substitution dict for commands etc., if any.
        wrapper (BackendWrapper): wrapper for command line
    Returns:
        Any: return value (e.g. exit code) of content
    """
    update_rlimits(backend.rlimits, log)

    venv = search = None
    if backend.native and backend.native.virtual_env:
        try:
            with substitutions_from(subst, raise_errors=True) as context:
                venv = context.evaluate(backend.native.virtual_env, 
                                        location=["backend.native.virtual_env"])
        except Exception as exc:
            raise BackendSpecificationError(f"error evaluating backend.native.virtual_env", exc)
        if venv:
            venv = os.path.expanduser(venv)
            if not os.path.isfile(f"{venv}/bin/activate"):
                raise BackendSpecificationError(f"virtual environment {venv} doesn't exist")
            log.debug(f"virtual environment is {venv}")

    args, log_args = build_command_line(cab, params, subst, virtual_env=venv, log=log)

    cabstat = cab.reset_status()

    command_name = cab.flavour.command_name

    # run command
    start_time = datetime.datetime.now()
    def elapsed(since=None):
        """Returns string representing elapsed time"""
        return str(datetime.datetime.now() - (since or start_time)).split('.', 1)[0]

    # log.info(f"argument lengths are {[len(a) for a in args]}")
    
    if wrapper:
        args, log_args = wrapper.wrap_run_command(args, log_args, fqname=fqname, log=log)
        
    log.debug(f"command line is {' '.join(log_args)}")

    retcode = xrun(args[0], args[1:], shell=False, log=log,
                output_wrangler=cabstat.apply_wranglers,
                return_errcode=True, command_name=command_name, 
                gentle_ctrl_c=True,
                log_command=' '.join(log_args), 
                log_result=False)

    # check if output marked it as a fail
    if cabstat.success is False:
        log.error(f"declaring '{command_name}' as failed based on its output")

    # if retcode != 0 and not explicitly marked as success, mark as failed
    if retcode and cabstat.success is not True:
        cabstat.declare_failure(f"{command_name} returns error code {retcode} after {elapsed()}")
    else:
        log.info(f"{command_name} returns exit code {retcode} after {elapsed()}")

    return cabstat




import sys

def example_function(x, y):
    print(f"x is {x}")
    print(f"y is {y}", file=sys.stderr)
    return 0

def example_function_dict(x, y):
    print(f"x is {x}")
    print(f"y is {y}", file=sys.stderr)
    return dict(a=x, b=y)



import click
from typing import List
from stimela.main import cli

import click
from typing import List, Optional, Tuple
from .run import run

from stimela.main import cli

@cli.command("build",
    help="""
    Builds singularity images required by the recipe. Only available if the singularity backend is selected.
    """,
    no_args_is_help=True)
@click.option("-r", "--rebuild", is_flag=True,
                help="""rebuilds all images from scratch. Default builds missing images only.""")
@click.option("-a", "--all-steps", is_flag=True,
                help="""builds images for all steps. Default is to omit explicitly skipped steps.""")
@click.option("-s", "--step", "step_ranges", metavar="STEP(s)", multiple=True,
                help="""only build images for specific step(s) from the recipe. Use commas, or give multiple times to cherry-pick steps.
                Use [BEGIN]:[END] to specify a range of steps. Note that cherry-picking an individual step via this option
                also impies --enable-step.""")
@click.option("-t", "--tags", "tags", metavar="TAG(s)", multiple=True,
                help="""only build images for steps wth the given tags (and also steps tagged as "always"). 
                Use commas, or give multiple times for multiple tags.""")
@click.option("--skip-tags", "skip_tags", metavar="TAG(s)", multiple=True,
                help="""explicitly skips steps wth the given tags. 
                Use commas, or give multiple times for multiple tags.""")
@click.option("-e", "--enable-step", "enable_steps", metavar="STEP(s)", multiple=True,
                help="""Build image for step(s) even if the recipe marks them as skipped. Use commas, or give multiple times 
                for multiple steps.""")
@click.option("-c", "--config", "config_equals", metavar="X.Y.Z=VALUE", nargs=1, multiple=True,
                help="""tweak configuration options.""")
@click.option("-C", "--config-assign", metavar="X.Y.Z VALUE", nargs=2, multiple=True,
                help="""tweak configuration options: same function -c/--config, but plays nicer with the shell's 
                tab completion feature.""")
@click.option("-l", "--last-recipe", is_flag=True,
                help="""if multiple recipes are defined, selects the last one for building.""")
@click.option("-S", "--singularity", "enable_singularity", is_flag=True,
                help="""Selects the singularity backend (shortcut for -C opts.backend.select=singularity)""")
@click.option("--slurm", "enable_slurm", is_flag=True,
                help="""Enables the slurm backend wrapper (shortcut for -C backend.slurm.enable=True)""")
@click.argument("what", metavar="filename.yml ... [recipe name] [PARAM=VALUE] ...", nargs=-1, required=True) 
def build(what: str, last_recipe: bool = False, rebuild: bool = False, all_steps: bool=False,
            config_equals: List[str] = [],
            config_assign: List[Tuple[str, str]] = [],
            step_ranges: List[str] = [], tags: List[str] = [], skip_tags: List[str] = [], enable_steps: List[str] = [],
            enable_singularity=False,
            enable_slurm=False):
    return run.callback(what, last_recipe=last_recipe, step_ranges=step_ranges, 
        tags=tags, skip_tags=skip_tags, enable_steps=enable_steps,
        config_equals=config_equals, config_assign=config_assign,
        build=True, rebuild=rebuild, build_skips=all_steps,
        enable_singularity=enable_singularity,
        enable_slurm=enable_slurm)



import logging
from typing import Dict, Optional, Any, List
from dataclasses import dataclass
from omegaconf import OmegaConf
import stimela
from stimela.backends import StimelaBackendOptions
from stimela.exceptions import BackendError

from . import get_backend

@dataclass
class BackendRunner(object):
    opts: StimelaBackendOptions
    is_remote: bool
    is_remote_fs: bool
    backend: Any
    backend_name: str
    wrapper: Any

    def run(self, cab: 'stimela.kitchen.cab.Cab', params: Dict[str, Any], fqname: str,
            log: logging.Logger, subst: Optional[Dict[str, Any]] = None):
        return self.backend.run(cab, params, fqname=fqname, backend=self.opts, log=log, subst=subst, 
                                wrapper=self.wrapper)
        
    def build(self, cab: 'stimela.kitchen.cab.Cab', log: logging.Logger, rebuild=False):
        if not hasattr(self.backend, 'build'):
            log.warning(f"the {self.backend_name} backend does support or require image builds")
        else:
            return self.backend.build(cab, backend=self.opts, log=log, rebuild=rebuild,
                                wrapper=self.wrapper)
    

def validate_backend_settings(backend_opts: Dict[str, Any], log: logging.Logger) -> BackendRunner:
    """Checks that backend settings refer to a valid backend
    
    Returs tuple of options, main, wrapper, where 'main' the the main backend, and 'wrapper' is an optional wrapper backend 
    such as slurm.
    """
    if not isinstance(backend_opts, StimelaBackendOptions):
        backend_opts = OmegaConf.to_object(backend_opts)

    backend_name = backend = None
    selected = backend_opts.select or ['singularity', 'native']
    # select containerization engine, if any
    for name in selected: 
        # check that backend has not been disabled
        opts = getattr(backend_opts, name, None)
        if not opts or opts.enable:
            backend = get_backend(name, opts)
            if backend is not None:
                backend_name = name
                break
    else:
        raise BackendError(f"selected backends ({', '.join(selected)}) not available")

    is_remote = is_remote_fs = backend.is_remote()

    # check if slurm wrapper is to be applied
    if backend_opts.slurm.enable:
        if is_remote:
            raise BackendError(f"can't combine slurm with {backend_name} backend")
        is_remote = True
        is_remote_fs = False
        backend_opts.slurm.validate(log)
        wrapper = backend_opts.slurm
    # otherwise use empty wrapper
    else:
        wrapper = None

    return BackendRunner(opts=backend_opts, is_remote=is_remote, is_remote_fs=is_remote_fs, 
                        backend=backend, backend_name=backend_name,
                        wrapper=wrapper)



import subprocess
import os
import logging
import pathlib
import shutil
from tempfile import mkdtemp
from contextlib import ExitStack
from enum import Enum
import stimela
from shutil import which
from dataclasses import dataclass
from omegaconf import OmegaConf
from typing import Dict, Any, Optional, Union
from scabha.basetypes import EmptyDictDefault
import datetime
from stimela.utils.xrun_asyncio import xrun
from stimela.exceptions import BackendError
from . import native

ReadWrite = Enum("BindMode", "ro rw", module=__name__)

@dataclass
class SingularityBackendOptions(object):
    @dataclass
    class BindDir(object):
        host: Optional[str] = None      # host path, default uses label, or else "empty" for tmpdir or "shm" for shm tmpdir
        target: Optional[str] = None    # container path: ==host by default
        mode: ReadWrite = "rw"
        mkdir: bool = False             # create host directory if it doesn't exist
        conditional: Union[bool, str] = True # bind conditionally (will be formula-evaluated)

    enable: bool = True
    image_dir: str = os.path.expanduser("~/.singularity")
    auto_build: bool = True
    rebuild: bool = False
    executable: Optional[str] = None
    remote_only: bool = False         # if True, won't look for singularity on local system -- useful in combination with slurm wrapper

    contain: bool = True              # if True, runs with --contain
    containall: bool = False          # if True, runs with --containall
    bind_tmp: Union[bool, str] = True # binds /tmp to "tmp" class if True. If string, uses specified ephemeral storage class (e.g. "ram")
    clean_tmp: bool = True            # if False, temporary directories will not be cleaned up. Useful for debugging.

    # optional extra bindings
    bind_dirs: Dict[str, BindDir] = EmptyDictDefault()
    # extra environment variable settings
    env: Dict[str, str] = EmptyDictDefault()

    # ephemeral storage settings, e.g. {tmp: /tmp, ram: /dev/shm}
    ephemeral: Dict[str, str] = EmptyDictDefault()


SingularityBackendSchema = OmegaConf.structured(SingularityBackendOptions)

STATUS = VERSION = BINARY = None

# images rebuilt in this run
_rebuilt_images = set()


class CustomTemporaryDirectory(object):
    """Custom context manager for tempfile.mkdtemp()."""
    def __init__(self, clean_up=True, dir=None):
        self.name = mkdtemp(dir=dir)
        self.clean_up = clean_up  # Workaround for < Python3.12

    def __enter__(self):
        return self.name

    def __exit__(self, exc_type, exc_value, traceback):
        if self.clean_up:
            shutil.rmtree(self.name)


def is_available(opts: Optional[SingularityBackendOptions] = None):
    global STATUS, VERSION, BINARY
    if STATUS is None:
        if opts and opts.remote_only:
            STATUS = VERSION = "remote"
        else:
            BINARY = (opts and opts.executable) or which("singularity")
            if BINARY:
                __version_string = subprocess.check_output([BINARY, "--version"]).decode("utf8")
                STATUS = VERSION = __version_string.strip().split()[-1]
                # if VERSION < "3.0.0":
                #     suffix = ".img"
                # else:
                #     suffix = ".sif"
            else:
                STATUS = "not installed"
                VERSION = None    
    return VERSION is not None

def get_status():
    is_available()
    return STATUS

def is_remote():
    return False

def init(backend: 'stimela.backend.StimelaBackendOptions', log: logging.Logger):
    pass

def get_image_info(cab: 'stimela.kitchen.cab.Cab', backend: 'stimela.backend.StimelaBackendOptions'):
    """returns image name/path corresponding to cab

    Args:
        cab (stimela.kitchen.cab.Cab): _description_
        backend (stimela.backend.StimelaBackendOptions): _description_

    Returns:
        name, path, enable_update: tuple of docker image name, path to singularity image on disk, and enable-updates flag
    """

    # prebuilt image
    if cab.image and cab.image.path:
        simg_path = cab.image.path
        if not os.path.exists(simg_path):
            raise BackendError(f"image {simg_path} for cab '{cab.name}' doesn't exist")
        return os.path.basename(simg_path), simg_path

    image_name = cab.flavour.get_image_name(cab, backend)

    if not image_name:
        raise BackendError(f"cab '{cab.name}' does not define an image")
    
    # convert to filename
    simg_name = image_name.replace("/", "-") + ".simg"
    simg_path = os.path.join(backend.singularity.image_dir, simg_name) 

    return image_name, simg_path


def build(cab: 'stimela.kitchen.cab.Cab', backend: 'stimela.backend.StimelaBackendOptions', log: logging.Logger,
            wrapper: Optional['stimela.backend.runner.BackendWrapper']=None,
            build=True, rebuild=False):
    """Builds image for cab, if necessary.

    build: if True, build missing images regardless of backend settings
    rebuild: if True, rebuild all images regardless of backend settings

    Returns:
        str: path to corresponding singularity image
    """

    # ensure image directory exists
    if os.path.exists(backend.singularity.image_dir):
        if not os.path.isdir(backend.singularity.image_dir):
            raise BackendError(f"invalid singularity image directory {backend.singularity.image_dir}")
    else:
        try:
            pathlib.Path(backend.singularity.image_dir).mkdir(parents=True)
        except OSError as exc:
            raise BackendError(f"failed to create singularity image directory {backend.singularity.image_dir}: {exc}")

    image_name, simg_path = get_image_info(cab, backend)

    # this is True if we're allowed to build missing images
    build = build or rebuild or backend.singularity.auto_build   
    # this is True if we're asked to force-rebuild images
    rebuild = rebuild or backend.singularity.rebuild
    
    cached_image_exists = os.path.exists(simg_path)

    # no image? Better have builds enabled then
    if not cached_image_exists:
        log.info(f"singularity image {simg_path} does not exist")
        if not build:
            raise BackendError(f"no image, and singularity build options not enabled")
    # else we have an image
    # if rebuild is enabled, delete it
    elif rebuild:
        if simg_path in _rebuilt_images:
            log.info(f"singularity image {simg_path} was rebuilt earlier")
        else:
            log.info(f"singularity image {simg_path} exists but a rebuild was specified")
            os.unlink(simg_path)        
            cached_image_exists = False
    else:
        log.info(f"singularity image {simg_path} exists")

    ## OMS: taking this out for now, need some better auto-update logic, let's come back to it later
    ## Please retain the code for now
        
    # # else check if it need to be auto-updated
    # elif auto_update_allowed and backend.singularity.auto_update:
    #     if image_name in _auto_updated_images:
    #         log.info("image was used earlier in this run, not checking for auto-updates again")
    #     else:
    #         _auto_updated_images.add(image_name)
    #         # force check of docker binary
    #         docker.is_available()
    #         if docker.BINARY is None:
    #             log.warn("a docker runtime is required for auto-update of singularity images: forcing unconditional rebuild")
    #             build = True
    #         else:
    #             log.info("singularity auto-update: pulling and inspecting docker image")
    #             # pull image from hub
    #             retcode = xrun(docker.BINARY, ["pull", image_name], 
    #                         shell=False, log=log,
    #                             return_errcode=True, command_name="(docker pull)", 
    #                             log_command=True, 
    #                             log_result=True)
    #             if retcode != 0:
    #                 raise BackendError(f"docker pull failed with return code {retcode}") 
    #             if os.path.exists(simg_path):
    #                 # check timestamp
    #                 result = subprocess.run(
    #                         [docker.BINARY, "inspect", "-f", "{{ .Created }}", image_name],
    #                         capture_output=True)
    #                 if result.returncode != 0:
    #                     for line in result.stdout.split("\n"):
    #                         log.warn(f"docker inpect stdout: {line}")
    #                     for line in result.stderr.split("\n"):
    #                         log.error(f"docker inpect stderr: {line}")
    #                     raise BackendError(f"docker inspect failed with return code {result.returncode}")
    #                 timestamp = result.stdout.decode().strip()
    #                 log.info(f"docker inspect returns timestamp {timestamp}")
    #                 # parse timestamps like '2023-04-07T13:39:19.187572398Z'
    #                 # Pre-3.11 pythons don't do it natively so we mess around...
    #                 match = re.fullmatch("(.*)T([^.]*)(\.\d+)?Z?", timestamp)
    #                 if not match:
    #                     raise BackendError(f"docker inspect returned invalid timestamp '{timestamp}'")
    #                 try:
    #                     dt = datetime.datetime.fromisoformat(f'{match.group(1)} {match.group(2)} +00:00')
    #                 except ValueError as exc:
    #                     raise BackendError(f"docker inspect returned invalid timestamp '{timestamp}', exc")

    #                 if dt.timestamp() > os.path.getmtime(simg_path):
    #                     log.warn("docker image is newer than cached singularity image, rebuilding")
    #                     os.unlink(simg_path)        
    #                     cached_image_exists = False
    #                 else:
    #                     log.info("cached singularity image appears to be up-to-date")

    # if image doesn't exist, build it. We will have already checked for build settings
    # being enabled above
    if not cached_image_exists:
        log.info(f"(re)building image {simg_path}")

        args = log_args = [BINARY, "build", simg_path, f"docker://{image_name}"]

        if wrapper:
            args, log_args = wrapper.wrap_build_command(args, log=log)

        retcode = xrun(args[0], args[1:], shell=False, log=log,
                    return_errcode=True, command_name="(singularity build)", 
                    gentle_ctrl_c=True,
                    log_command=' '.join(log_args), 
                    log_result=True)

        if retcode:
            raise BackendError(f"singularity build returns {retcode}")

        if not os.path.exists(simg_path):
            raise BackendError(f"singularity build did not return an error code, but the image did not appear")
        
        _rebuilt_images.add(simg_path)
        
    return simg_path



def run(cab: 'stimela.kitchen.cab.Cab', params: Dict[str, Any], fqname: str,
        backend: 'stimela.backend.StimelaBackendOptions',
        log: logging.Logger, subst: Optional[Dict[str, Any]] = None,
        wrapper: Optional['stimela.backends.runner.BackendWrapper'] = None):

    """Runs cab contents

    Args:
        cab (Cab): cab object
        log (logger): logger to use
        subst (Optional[Dict[str, Any]]): Substitution dict for commands etc., if any.

    Returns:
        Any: return value (e.g. exit code) of content
    """
    from .utils import resolve_required_mounts

    native.update_rlimits(backend.rlimits, log)

    # get path to image, rebuilding if backend options allow this
    simg_path = build(cab, backend=backend, log=log, build=False, wrapper=wrapper)

    # build up command line    
    cwd = os.getcwd()
    args = [backend.singularity.executable or BINARY, 
            "exec", 
            "--pwd", cwd]
    if backend.singularity.containall:
        args.append("--containall")
    elif backend.singularity.contain:
        args.append("--contain")
    if backend.singularity.env:
        args += ["--env", ",".join([f"{k}={v}" for k, v in backend.singularity.env.items()])]

    # initial set of mounts has cwd as read-write
    mounts = {cwd: True}
    # dict of container paths to host paths
    container_to_host_path = {}

    # get dict of ephemeral base directories
    ephem_classes = backend.singularity.ephemeral
    if not ephem_classes:
        ephem_classes["tmp"] = "/tmp"
    ephem_binds = {}

    with ExitStack() as exit_stack: 
        # add extra binds
        for label, bind in backend.singularity.bind_dirs.items():
            # skip if conditional is False
            if not bind.conditional:
                log.info(f"bind_dirs.{label}: skipping based on conditional == {bind.conditional}")
                continue

            # expand ~ in paths
            src = os.path.expanduser(bind.host).rstrip("/")
            dest = os.path.expanduser(bind.target or src).rstrip("/")
            rw = bind.mode == ReadWrite.rw

            # handle binding of ephemeral dirs
            if bind.host == "empty":
                if "tmp" not in ephem_classes:
                    raise BackendError(f"bind_dirs.{label}: deprecated 'host: empty' setting requires that a 'tmp' ephemeral storage class be defined")
                bind.host = "ephemeral:tmp"
            ephem_class = bind.host.startswith("ephemeral:")
            if ephem_class or bind.host == "ephemeral":
                if not bind.target:
                    raise BackendError(f"bind_dirs.{label}: a target must be specified when host=ephemeral")
                target = os.path.expanduser(bind.target)
                # parse storage class specification, if given
                if ephem_class:
                    ephem_class = bind.host.split(':', 1)[1]
                    if ephem_class not in ephem_classes:
                        raise BackendError(f"bind_dirs.{label}: class '{ephem_class}' not present in 'singularity.ephemeral' section")
                    else:
                        ephem_target = ephem_classes[ephem_class]
                else:
                    ephem_class, ephem_target = list(ephem_classes.items())[0]
                # create temporary directory, or else stash ephem binding for the wrapper
                if wrapper:
                    src = f"EPH{len(ephem_binds)}"
                    ephem_binds[src] = ephem_target
                    src = f"::{src}::"
                else:
                    tmpdir = CustomTemporaryDirectory(clean_up=backend.singularity.clean_tmp, dir=ephem_target)
                    src = exit_stack.enter_context(tmpdir)
                    log.info(f"bind_dirs.{label}: using temporary directory {src}")
                container_to_host_path[target] = src
                mounts[src] = rw
            # else real binding
            else:
                # resolve symlinks
                if os.path.exists(src) and os.path.realpath(src) != src:
                    src = os.path.realpath(src)
                    log.info(f"bind_dirs.{label}: binding symlink target {src}")
                
                # make directory if needed
                if bind.mkdir:
                    # I think files can be bound too, so only do this check for directories
                    if os.path.exists(src):
                        if not os.path.isdir(src):
                            raise BackendError(f"bind_dirs.{label}: host path is not a directory")
                    else:
                        try:
                            pathlib.Path(src).mkdir(parents=True)
                        except Exception as exc:
                            raise BackendError(f"bind_dirs.{label}: error creating directory {bind.host}", exc)
                elif not os.path.exists(src):
                    raise BackendError(f"bind_dirs.{label}: host path '{src}' does not exist")
                
                # if already present in mounts, potentially upgrade to rw
                mounts[src] = mounts.get(src, False) or rw
                # if paths different, create a remapping
                if src != dest:
                    if dest in container_to_host_path:
                        if container_to_host_path[dest] != src:
                            raise BackendError(f"bind_dirs.{label}: conflicting bind paths for {dest}")
                    else:
                        container_to_host_path[dest] = src

        # get extra required filesystem bindings from supplied parameters
        resolve_required_mounts(mounts, params, cab.inputs, cab.outputs, remappings=container_to_host_path)

        # redo mounts as a list of (container_path, source_path, rw)
        source_to_container_path = {src: dest for dest, src in container_to_host_path.items()}
        # make list of mounts
        mounts = [(source_to_container_path.get(src, src), src, rw) for src, rw in mounts.items()]
        # add implicit /tmp mount
        if backend.singularity.bind_tmp:
            for target, _, _ in mounts:
                if target == "/tmp":
                    log.info("/tmp directory already bound, not adding an explicit binding")
                    break
            else:
                if type(backend.singularity.bind_tmp) is str:
                    tmp_class = backend.singularity.bind_tmp
                else:
                    tmp_class = "tmp"
                tmp_target = ephem_classes.get(tmp_class, None)
                if tmp_target is None:
                    raise BackendError(f"bind_tmp uses an undefined ephemeral storage class '{tmp_class}'")
                if wrapper:
                    src = f"EPH{len(ephem_binds)}"
                    ephem_binds[src] = tmp_target
                    mounts.append(("/tmp", f"::{src}::", True))
                else:
                    tmpdir = CustomTemporaryDirectory(clean_up=backend.singularity.clean_tmp, dir=tmp_target)
                    tmpdir_name = exit_stack.enter_context(tmpdir)
                    mounts.append(("/tmp", tmpdir_name, True))

        # sort mount paths before iterating -- this ensures that parent directories come first
        # (singularity doesn't like it if you specify a bind of a subdir before a bind of a parent) 
        for dest, src, rw in sorted(mounts):
            mode = 'rw' if rw else 'ro'
            if src == dest:
                log.info(f"binding {src} as {mode}")
            else:
                log.info(f"binding {src} to {dest} as {mode}")
            args += ["--bind", f"{src}:{dest}:{mode}"]

        args += [simg_path]
        log_args = args.copy()

        args1, log_args1 = cab.flavour.get_arguments(cab, params, subst, check_executable=False, log=log)
        args += args1
        log_args += log_args1

        cabstat = cab.reset_status()

        command_name = f"{cab.flavour.command_name}" or None

        # run command
        start_time = datetime.datetime.now()
        def elapsed(since=None):
            """Returns string representing elapsed time"""
            return str(datetime.datetime.now() - (since or start_time)).split('.', 1)[0]

        # log.info(f"argument lengths are {[len(a) for a in args]}")

        if wrapper:
            args, log_args = wrapper.wrap_run_command(args, log_args, ephem_binds=ephem_binds, fqname=fqname, log=log)

        log.debug(f"command line is {' '.join(log_args)}")

        retcode = xrun(args[0], args[1:], shell=False, log=log,
                    output_wrangler=cabstat.apply_wranglers,
                    return_errcode=True, command_name=command_name, 
                    gentle_ctrl_c=True,
                    log_command=' '.join(log_args), 
                    log_result=False)

        # check if output marked it as a fail
        if cabstat.success is False:
            log.error(f"declaring '{command_name}' as failed based on its output")

        # if retcode != 0 and not explicitly marked as success, mark as failed
        if retcode and cabstat.success is not True:
            cabstat.declare_failure(f"{command_name} returns error code {retcode} after {elapsed()}")
        else:
            log.info(f"{command_name} returns exit code {retcode} after {elapsed()}")

        return cabstat


# class SingularityError(Exception):
#     pass

# def pull(image, name, docker=True, directory=".", force=False):
#     """ 
#         pull an image
#     """
#     if docker:
#         fp = "docker://{0:s}".format(image)
#     else:
#         fp = image
#     if not os.path.exists(directory):
#         os.mkdir(directory)

#     image_path = os.path.abspath(os.path.join(directory, name))
#     if os.path.exists(image_path) and not force:
#         stimela.logger().info(f"Singularity image already exists at '{image_path}'. To replace it, please re-run with the 'force' option")
#     else:
#         utils.xrun(f"cd {directory} && singularity", ["pull", 
#         	"--force" if force else "", "--name", 
#          	name, fp])

#     return 0

# class Container(object):
#     def __init__(self, image, name,
#                  volumes=None,
#                  logger=None,
#                  time_out=-1,
#                  runscript="/singularity",
#                  environs=None,
#                  workdir=None,
#                  execdir="."):
#         """
#         Python wrapper to singularity tools for managing containers.
#         """

#         self.image = image
#         self.volumes = volumes or []
#         self.environs = environs or []
#         self.logger = logger
#         self.status = None
#         self.WORKDIR = workdir
#         self.RUNSCRIPT = runscript
#         self.PID = os.getpid()
#         self.uptime = "00:00:00"
#         self.time_out = time_out
#         self.execdir = execdir

#         self._env = os.environ.copy()

#         hashname = hashlib.md5(name.encode('utf-8')).hexdigest()[:3]
#         self.name = hashname if version < "3.0.0" else name

#     def add_volume(self, host, container, perm="rw", noverify=False):
#         if os.path.exists(host) or noverify:
#             if self.logger:
#                 self.logger.debug("Mounting volume [{0}] in container [{1}] at [{2}]".format(
#                     host, self.name, container))
#             host = os.path.abspath(host)
#         else:
#             raise IOError(
#                 "Path {0} cannot be mounted on container: File doesn't exist".format(host))

#         self.volumes.append(":".join([host, container, perm]))

#         return 0

#     def add_environ(self, key, value):
#         self.logger.debug("Adding environ varaible [{0}={1}] "\
#                     "in container {2}".format(key, value, self.name))
#         self.environs.append("=".join([key, value]))
#         key_ = f"SINGULARITYENV_{key}"
	
#         self.logger.debug(f"Setting singularity environmental variable {key_}={value} on host")
#         self._env[key_] = value

#         return 0

#     def run(self, *args, output_wrangler=None):
#         """
#         Run a singularity container instance
#         """

#         if self.volumes:
#             volumes = " --bind " + " --bind ".join(self.volumes)
#         else:
#             volumes = ""

#         if not os.path.exists(self.image):
#             self.logger.error(f"The image, {self.image}, required to run this cab does not exist."\
#                     " Please run 'stimela pull --help' for help on how to download the image")
#             raise SystemExit from None

#         self.status = "running"
#         self._print("Starting container [{0:s}]. Timeout set to {1:d}. The container ID is printed below.".format(
#             self.name, self.time_out))
        
#         utils.xrun(f"cd {self.execdir} && singularity", ["run", "--workdir", self.execdir, "--containall"] \

#                     + list(args) + [volumes, self.image, self.RUNSCRIPT],
#                     log=self.logger, timeout=self.time_out, output_wrangler=output_wrangler,
#                     env=self._env, logfile=self.logfile)

#         self.status = "exited"

#         return 0

#     def _print(self, message):
#         if self.logger:
#             self.logger.info(message)
#         else:
#             print(message)

#         return 0



import os
import re
import click
from scabha.exceptions import SchemaError
from .cargo import Parameter, UNSET, _UNSET_DEFAULT, Cargo, ParameterPolicies
from typing import List, Union, Optional, Callable, Dict, DefaultDict, Any
from .basetypes import EmptyDictDefault, File, is_file_type
from dataclasses import dataclass, make_dataclass, field, asdict
from omegaconf import OmegaConf, MISSING
from collections.abc import MutableSet, MutableSequence, MutableMapping
from scabha import configuratt
from collections import OrderedDict

def schema_to_dataclass(io: Dict[str, Parameter], class_name: str, bases=(), post_init: Optional[Callable] =None):
    """
    Converts a scabha schema to a dataclass.
    Each parameter in the schema will correspond to a field. Metadata of fields will contain:
    'help' for info, 'choices' for the choices field, 'parameter' for parameter name

    Args:
        io (Dict[str, Parameter]): dict of parameters
        class_name (str):           name of dataclass
        bases (tuple, optional): [description]. Base classes (passed to bases of make_dataclass())
        post_init (callable, optional): Inserts a __post__init__ method into the class if needed.

    Raises:
        SchemaError: if any parameters are malformed

    Returns:
        dataclass
    """
    field2name = {}
    fields = []
    for name, schema in io.items():
        if type(schema) is not Parameter:
            schema = Parameter(**schema)

        if is_file_type(schema._dtype):
            schema._dtype = str
            schema.dtype = "str"

        # sanitize name: dataclass won't take hyphens or periods
        # so replace with "_" and ensure uniqueness
        fldname = name.replace("-", "_").replace(".", "_")
        while fldname in field2name:
            fldname += "_"
        field2name[fldname] = name

        # form up metadata
        metadata = dict(help=schema.info, parameter=name)
        metadata.update(schema.metadata)
        if schema.choices:
            metadata['choices'] = schema.choices
        if schema.element_choices:
            metadata['element_choices'] = schema.element_choices
        metadata['required'] = required = schema.required

        if required and schema.default is not UNSET:
            raise SchemaError(
                f"Field '{fldname}' is required but specifies a default. "
                f"This behaviour is unsupported/ambiguous."
            )

        if required:
            fld = field(default=MISSING, metadata=metadata)
        elif isinstance(schema.default, MutableSequence):
            fld = field(default_factory=default_wrapper(list, schema.default),
                        metadata=metadata)
        elif isinstance(schema.default, MutableSet):
            fld = field(default_factory=default_wrapper(set, schema.default),
                        metadata=metadata)
        elif isinstance(schema.default, MutableMapping):
            fld = field(default_factory=default_wrapper(dict, schema.default),
                        metadata=metadata)
        elif schema.default is UNSET:
            fld = field(default=None, metadata=metadata)
        else:
            fld = field(default=schema.default, metadata=metadata)

        fields.append((fldname, schema._dtype, fld))

    namespace = None if post_init is None else dict(__post_init__=post_init)

    return make_dataclass(class_name, fields, bases=bases, namespace=namespace)


def default_wrapper(default_type, default_value):

    def default_factory():
        return default_type(default_value)

    return default_factory


def nested_schema_to_dataclass(nested: Dict[str, Dict], class_name: str, bases=(), section_bases=(), post_init_map={}):
    """Converts a nested schema (sections consisting of parameters) into a dataclass.

    Args:
        nested (Dict[str, Dict]):           schema sections
        class_name (str):                   name of dataclass created
        bases (tuple, optional):            base classes for outer class
        section_bases (tuple, optional):    base class for section classes
        post_init_map (dict, optional):     dict of specific __post_init__ methods to be attached to section classes

    Returns:
        tuple of (dataclass, dict):         resulting dataclass, plus dict of dataclasses for the sections
    """

    # make dataclass based on schema contents
    nested_structure = make_dataclass(f"_{class_name}_schemas", [(name, Dict[str, Parameter]) for name in nested.keys()])
    # turn into a structured config
    nested = OmegaConf.unsafe_merge(OmegaConf.structured(nested_structure), nested)
    fields = []

    # convert per-section schemas into dataclasses and make list of fields for outer dataclass
    for section, content in nested.items():
        dcls = schema_to_dataclass(content, f"{class_name}_{section}",
                                    bases=section_bases, post_init=post_init_map.get(section))

        fields.append((section, dcls, field(default_factory=dcls)))

    # return the outer dataclass
    return make_dataclass(class_name, fields, bases=bases)


_atomic_types = dict(bool=bool, str=str, int=int, float=float)

def _validate_list(text: str, element_type, schema, sep=",", brackets=True):
    if not text:
        if schema.default in (UNSET, _UNSET_DEFAULT):
            return None
        else:
            return schema.default

    if text == "[]":
        return []

    if text[0] == "[" and text[-1] == "]":
        text = text[1:-1]
    elif brackets:
        raise click.BadParameter(f"can't convert to '{schema.dtype}', missing '[]' brackets")

    try:
        return list(element_type(x) for x in text.split(sep))
    except ValueError as e:
        raise click.BadParameter(f"can't convert to '{schema.dtype}'. Underlying exception was: {e}")

def _validate_tuple(text: str, element_types, schema, sep=",", brackets=True):
    if not text:
        if schema.default in (UNSET, _UNSET_DEFAULT):
            return None
        else:
            return schema.default

    if text == "[]":
        return []

    if text[0] == "[" and text[-1] == "]":
        text = text[1:-1]
    elif brackets:
        raise click.BadParameter(f"can't convert to '{schema.dtype}', missing '[]' brackets")

    elems = text.split(sep)
    if len(elems) != len(element_types):
        raise click.BadParameter(f"can't convert to '{schema.dtype}', tuple length mismatch")
    try:
        return tuple(element_type(x) for x, element_type in zip(elems, element_types))
    except ValueError as e:
        raise click.BadParameter(f"can't convert to '{schema.dtype}'. Underlying exception was: {e}")

@dataclass
class Schema(object):
    inputs: Dict[str, Parameter] = EmptyDictDefault()
    outputs: Dict[str, Parameter] = EmptyDictDefault()
    policies: Optional[Dict[str, Any]] = None


def clickify_parameters(schemas: Union[str, Dict[str, Any]],
                        default_policies: Dict[str, Any] = None):
    """
    Create command line parameters from a YAML schema. Uses the click
    package.

    Args:
        schemas (str or Dict): Either the YAML filename from which the parameter schema is loaded,
            containing inputs, outputs and an [optional] policies section,
            or a DictConfig object containing inputs/outputs/policies.
            See https://stimela.readthedocs.io/en/latest/reference/schema_ref.html
        default_policies: default policies applied to the schema, overrides the policies section
            if supplied. See ParameterPolicies in scabha/cargo.py, and
            https://stimela.readthedocs.io/en/latest/reference/policies.html

    Example:
    =======
    The following code defines a simple function that takes two parameters.
    The schema is contained in the file hello/hello.yml. It contains

        inputs:
            count:
                dtype: int
                default: 1
            name:
                dtype: str
                required: true

        outputs:
        {}

    The corresponding python code uses clickify_parameters as
    a decorator:
        import click

        from scabha.schema_utils import clickify_parameters
        from omegaconf import OmegaConf


        @click.command()
        @clickify_parameters('hello/hello.yml')
        def hello(count: int = 1, name: Optional[str] = None):
            for x in range(count):
                click.echo(f"Hello {name}!")
    Returns:
        Nothing
    """

    if type(schemas) is str:
        schemas = OmegaConf.merge(OmegaConf.structured(Schema),
                                  OmegaConf.load(schemas))

    # get default policies from argument or schemas
    if default_policies:
        default_policies = OmegaConf.merge(OmegaConf.structured(ParameterPolicies), default_policies)
    elif getattr(schemas, 'policies', None):
        default_policies = OmegaConf.merge(OmegaConf.structured(ParameterPolicies), schemas.policies)
    else:
        default_policies = ParameterPolicies()

    decorator_chain = None
    inputs = Cargo.flatten_schemas(OrderedDict(), getattr(schemas, 'inputs', {}), "inputs")
    outputs = Cargo.flatten_schemas(OrderedDict(), getattr(schemas, 'outputs', {}), "outputs")
    for io in inputs, outputs:
        for name, schema in io.items():
            # skip outputs, unless they're named outputs
            if io is outputs and not (schema.is_file_type and not schema.implicit):
                continue

            # sometimes required to convert ParameterPolicies object to dict
            if isinstance(schema.policies, ParameterPolicies):
                merge_policies = {k: v for k, v in asdict(schema.policies).items() if v is not None}
            else:
                merge_policies = {k: v for k, v in schema.policies.items() if v is not None}
            # None should not take precedence in the merge
            policies = OmegaConf.merge(default_policies, merge_policies)

            # impose default repeat policy of using a single argument for a list, i.e. X1,X2,X3
            if policies.repeat is None:
                policies.repeat = ","

            name = name.replace("_", "-").replace(".", "-")
            optname = f"--{name}"
            dtype = schema.dtype
            if type(dtype) is str:
                dtype = dtype.strip()
            validator = None
            multiple = False
            nargs = 1
            # process optional
            optional_match = re.fullmatch(r"Optional\[(.*)\]", dtype)
            if optional_match:
                str_dtype = dtype = optional_match.group(1).strip()
            else:
                str_dtype = str(dtype)
            metavar = schema.metavar or str_dtype
            is_unset = schema.default in (UNSET, _UNSET_DEFAULT)
            kwargs = dict()
            if not is_unset and not schema.suppress_cli_default and nargs != -1:
                kwargs['default'] = schema.default

            # sort out option type. Atomic type?
            if dtype in _atomic_types:
                dtype = _atomic_types[dtype]
                if dtype is bool:
                    optname = f"{optname}/--no-{name}"
            # file type? NB: URI not included deliberately -- this becomes a str in the else: clause below
            elif dtype in ("MS", "File", "Directory"):
                dtype = click.Path(exists=(io is schemas.inputs))
            else:
                list_match = re.fullmatch(r"List\[(.*)\]", dtype)
                tuple_match = re.fullmatch(r"Tuple\[(.*)\]", dtype)
                # List[x] type? Add validation callback to convert elements
                if list_match:
                    elem_type = _atomic_types.get(list_match.group(1).strip(), str)
                    if policies.repeat == 'list':
                        if not policies.positional:
                            raise SchemaError(f"click parameter '{name}': repeat=list policy is only supported for positional=true parameters")
                        nargs = -1
                        dtype = elem_type
                        metavar = schema.metavar or f"{elem_type.__name__} ..."
                    elif policies.repeat == 'repeat':
                        multiple = True
                        dtype = elem_type
                        metavar = schema.metavar or f"{elem_type.__name__}"
                        # multiple options have a click default of () not None
                        if 'default' not in kwargs:
                            kwargs['default'] = None
                    elif policies.repeat == '[]':  # else assume [X,Y] or X,Y syntax
                        dtype = str
                        validator = lambda ctx, param, value, etype=dtype, schema=schema, _type=elem_type: \
                            _validate_list(value, element_type=_type, schema=schema,
                                           brackets=False)
                        metavar = schema.metavar or f"{elem_type.__name__},{elem_type.__name__},..."
                    elif policies.repeat is not None:  # assume XrepY syntax
                        dtype = str
                        sep = policies.repeat
                        validator = lambda ctx, param, value, etype=dtype, schema=schema, _type=elem_type: \
                            _validate_list(value, element_type=_type, schema=schema,
                                           sep=sep, brackets=False)
                        metavar = schema.metavar or f"{elem_type.__name__}{sep}{elem_type.__name__}{sep}..."
                    else:
                        raise SchemaError(f"list-type parameter '{name}' does not have a repeat policy set")
                elif tuple_match:
                    elem_types = tuple(_atomic_types.get(t.strip(), str) for t in tuple_match.group(1).split(","))
                    if policies.repeat == 'list':
                        nargs = len(elem_types)
                        dtype = elem_types
                        metavar = schema.metavar or " ".join(t.__name__ for t in elem_types)
                    elif policies.repeat == 'repeat':
                        raise SchemaError(f"tuple-type parameter '{name}' has unsupported repeat policy 'repeat'")
                    elif policies.repeat == '[]':  # else assume [X,Y] or X,Y syntax
                        dtype = str
                        metavar = schema.metavar or ",".join((t.__name__ for t in elem_types))
                        validator = lambda ctx, param, value, etype=dtype, \
                                schema=schema, _type=elem_types: \
                                _validate_tuple(value, element_types=_type,
                                                schema=schema, brackets=False)
                    elif policies.repeat is not None:  # assume XrepY syntax
                        dtype = str
                        metavar = schema.metavar or policies.repeat.join((t.__name__ for t in elem_types))
                        validator = lambda ctx, param, value, etype=dtype, \
                        schema=schema, _type=elem_types: \
                            _validate_tuple(value, element_types=_type,
                                            schema=schema,
                                            sep=policies.repeat, brackets=False)
                    else:
                        raise SchemaError(f"tuple-type parameter '{name}' does not have a repeat policy set")
                else:
                    # anything else will be just a string
                    dtype = str

            # choices?
            if schema.choices:
                dtype = click.Choice(schema.choices)

            # aliases?
            optnames = [optname]
            if schema.abbreviation:
                optnames.append(f"-{schema.abbreviation}")

            if policies.positional:
                kwargs.update(type=dtype, callback=validator, required=schema.required, nargs=nargs,
                              metavar=metavar)
                deco = click.argument(name, **kwargs)
            else:
                kwargs.update(type=dtype, callback=validator,
                              required=schema.required, multiple=multiple,
                              metavar=metavar, help=schema.info)
                deco = click.option(*optnames, **kwargs)
            if decorator_chain is None:
                decorator_chain = deco
            else:
                decorator_chain = lambda x,deco=deco,chain=decorator_chain: chain(deco(x))

    return decorator_chain or (lambda x: x)

@dataclass
class SchemaSpec:
    inputs: Dict[str, Parameter]
    outputs: Dict[str, Parameter]
    libs: Dict[str, Any]

def paramfile_loader(paramfiles: Union[File, List[File]], sources: Union[File, List[File]] = [],
                     schema_spec=None, use_cache=False) -> Dict:
    """Load a scabha-style parameter defintion using.

    Args:
        paramfiles (List[File]): Name of parameter definition files
        sources (List[Dict], optional): Parameter definition dependencies
        (a.k.a files specified via_include)

    Returns:
        Dict: Schema object
    """
    args_defn = OmegaConf.structured(schema_spec or SchemaSpec)
    if isinstance(paramfiles, File):
        paramfiles = [paramfiles]

    if isinstance(sources, File):
        sources = [sources]

    srcs = []
    for src in sources:
        if not src.EXISTS:
            raise FileNotFoundError(f"Source file for either of {paramfiles} could not be found at {src.PATH}")
        srcs.append(configuratt.load(src, use_cache=use_cache)[0])

    struct_args, _ = configuratt.load_nested(paramfiles, structured=args_defn,
                                            use_sources=srcs, use_cache=use_cache)

    return OmegaConf.create(struct_args)




import fnmatch
import os, sys
import click
from typing import *
from rich.tree import Tree
from rich.table import Table
from rich import box
from rich import print as rich_print
from omegaconf import OmegaConf

import stimela
from stimela import logger, log_exception
from stimela.main import cli
from scabha.cargo import ParameterCategory
from stimela.kitchen.recipe import Recipe
from stimela.kitchen.cab import Cab
from stimela.exceptions import RecipeValidationError, CabValidationError
from stimela.task_stats import destroy_progress_bar

from .run import load_recipe_files, resolve_recipe_files

@cli.command("doc",
    help="""
    Print documentation on a cab or a recipe.
    """)
@click.option("do_list", "-l", "--list", is_flag=True,
                help="""Lists the available cabs and recipes, including custom-defined ones.""")
@click.option("-I", "--implicit", is_flag=True,
                help="""Increases level of detail to include implicit inputs/outputs.""")
@click.option("-O", "--obscure", is_flag=True,
                help="""Increases level of detail to include implicit and obscure inputs/outputs.""")
@click.option("-A", "--all", is_flag=True,
                help="""Increases level of detail to include all inputs/outputs.""")
@click.option("-R", "--required", is_flag=True,
                help="""Decreases level of detail to include required inputs/outputs only.""")
@click.argument("what", nargs=-1, metavar="filename.yml ... [recipe name] [cab name]", required=True) 
def doc(what: List[str] = [], do_list=False, implicit=False, obscure=False, all=False, required=False):

    log = logger()
    top_tree = Tree(f"stimela doc {' '.join(what)}", guide_style="dim")

    if required:
        max_category = ParameterCategory.Required
    else:
        max_category = ParameterCategory.Optional
    if all:
        max_category = ParameterCategory.Hidden
    elif obscure:
        max_category = ParameterCategory.Obscure
    elif implicit:
        max_category = ParameterCategory.Implicit
        
    def load_recipe(name: str, section: Dict):
        try:
            if not section.get('name'):
                section.name = name
            recipe = Recipe(**section)
            recipe.finalize(fqname=name)
            return recipe
        except Exception as exc:
            if not isinstance(exc, RecipeValidationError):
                exc = RecipeValidationError(f"error loading recipe '{name}'", exc)
            log_exception(exc)
            sys.exit(2)

    # load all recipe/config files
    files_to_load = []
    names_to_document = []
    errcode = None
    for item in what:
        try:
            filenames = resolve_recipe_files(item, log=log)
        except FileNotFoundError as exc:
            log_exception(exc)
            errcode = 2
            continue
        if filenames is not None:
            files_to_load += filenames
            log.info(f"will load recipe/config file(s) {', '.join(filenames)}")
        else:
            log.info(f"treating '{item}' as a recipe or cab name")
            names_to_document.append(item)
    
    # load config and recipes from all given files
    if files_to_load:
        top_level_recipes = load_recipe_files(files_to_load)

    destroy_progress_bar()

    if not files_to_load:
        log.error(f"No YAML documents were specified")
        sys.exit(2)

    if errcode:
        sys.exit(errcode)
    
    log.info(f"loaded {len(stimela.CONFIG.cabs)} cab definition(s) and {len(stimela.CONFIG.lib.recipes)} recipe(s)")

    if not stimela.CONFIG.lib.recipes and not stimela.CONFIG.cabs:
        log.error(f"Loaded YAML documents do not contain any cab definitions or recipes")
        sys.exit(2)

    recipes_to_document = []
    cabs_to_document = []

    for item in names_to_document:
        recipe_names = fnmatch.filter(stimela.CONFIG.lib.recipes.keys(), item)
        cab_names = fnmatch.filter(sorted(stimela.CONFIG.cabs.keys()), item)
        if not recipe_names and not cab_names:
            log.error(f"'{item}' does not match any recipe or cab names. Try -l/--list")
            sys.exit(2)
        recipes_to_document += recipe_names
        cabs_to_document += cab_names

    # if nothing was specified, and only one cab/only one recipe is defined, print that
    if not names_to_document:
        if len(top_level_recipes) == 1:
            recipes_to_document = top_level_recipes
            log.info("a single top-level recipe is defined, documenting it by default. Use -l to list all defined recipes/cabs")
        elif len(stimela.CONFIG.cabs) == 1 and not top_level_recipes:
            cabs_to_document = list(stimela.CONFIG.cabs.keys())

    if recipes_to_document or cabs_to_document:
        for name in recipes_to_document:
            recipe = load_recipe(name, stimela.CONFIG.lib.recipes[name])
            tree = top_tree.add(f"Recipe: [bold]{name}[/bold]")
            recipe.rich_help(tree, max_category=max_category)
        
        for name in cabs_to_document:
            try:
                cab = Cab(**stimela.CONFIG.cabs[name])
                cab.finalize(config=stimela.CONFIG)
            except Exception as exc:
                if not isinstance(exc, CabValidationError):
                    exc = CabValidationError(f"error loading cab '{name}'", exc)
                log_exception(exc)
                sys.exit(2)
            tree = top_tree.add(f"Cab: [bold]{name}[/bold]")
            cab.rich_help(tree, max_category=max_category)

    # list recipes and cabs -- also do this by default if nothing explicit was documented    
    if do_list or not (recipes_to_document or cabs_to_document):
        if stimela.CONFIG.lib.recipes:
            subtree = top_tree.add("Recipes:")
            table = Table.grid("", "", padding=(0,2))
            for name, recipe in stimela.CONFIG.lib.recipes.items():
                table.add_row(f"[bold]{name}[/bold]", getattr(recipe, 'info', ''))
            subtree.add(table)

        if stimela.CONFIG.cabs:
            subtree = top_tree.add("Cabs:")
            table = Table.grid("", "", padding=(0,2))
            for name, cab in sorted(stimela.CONFIG.cabs.items()):
                table.add_row(f"[bold]{name}[/bold]", getattr(cab, 'info', ''))
            subtree.add(table)            
        
    rich_print(top_tree)





import click
from typing import List
from stimela.main import cli
import stimela

@cli.command(
    help="""Push stimela base images. Specify a list of image names and versions (if a version is omitted, pushes all known versions of image),
            or else use --all to push everything. 
         """,
    short_help="push stimela images to registry",
    no_args_is_help=True)
@click.argument("images", nargs=-1, metavar="NAME[:VERSION]", required=False) 
@click.option("-a", "--all", is_flag=True, 
                help="push all images")
def push(images: List[str], all=False):
    from stimela.main import BACKEND
    from stimela import CONFIG
    log = stimela.logger()

    available_images = BACKEND.available_images()
    push_images = CONFIG.base.keys() if all else images

    for imagename in push_images:
        if ':' in imagename:
            imagename, version = imagename.split(":", 1)
        else:
            version = None

        if imagename not in CONFIG.base:
            log.error(f"base image '{imagename}' is not known to Stimela")
            return 2

        image = CONFIG.base[imagename]

        if version is None:
            push_versions = image.images.keys()
        elif version in image.images:
            push_versions = [version]
        else:
            log.error(f"version '{version}' is not defined for base image '{imagename}'")
            return 2

        # now loop over push versions
        for version in push_versions:
            # check if already exists
            if imagename not in available_images or version not in available_images[imagename] \
                        or available_images[imagename][version].build is None:
                log.warning(f"image '{imagename}:{version}' not found, has it been built?")
                continue

            BACKEND.push(image, version)





from dataclasses import dataclass
from typing import Optional
import logging
import stimela

# add these as module attributes
from .run_native import run, build_command_line, update_rlimits

def is_available(opts = None):
    return True

def get_status():
    return "OK"

def is_remote():
    return False


@dataclass
class NativeBackendOptions(object):
    enable: bool = True
    virtual_env: Optional[str] = None
    

def init(backend: 'stimela.backend.StimelaBackendOptions', log: logging.Logger):
    pass



import logging
from typing import Dict, Any
import re
from datetime import datetime
from rich.markup import escape
from requests import ConnectionError
from urllib3.exceptions import HTTPError

from omegaconf import OmegaConf

from stimela.exceptions import StimelaCabRuntimeError

from kubernetes.client.rest import ApiException
from . import get_kube_api, KubeBackendOptions

k8s_cpu_units = {
    "": 1,
    "m": 1e-3,
    "u": 1e-6, # not sure this exists
    "n": 1e-9
}

k8s_memory_units_in_bytes = {
    "": 1,
    "b": 1,
    "Ki": 2**10,
    "Mi": 2**20,
    "Gi": 2**30,
    "Ti": 2**40,
    "Pi": 2**50,
    "Ei": 2**60,
    # Decimal units
    "K": 10**3,
    "M": 10**6,
    "G": 10**9,
    "T": 10**12,
    "P": 10**15,
    "E": 10**18
}

def resolve_unit(quantity:str, units: Dict = k8s_memory_units_in_bytes):
    match = re.fullmatch(r"^(\d+)(.*)$", quantity)
    if not match or match.group(2) not in units:
        raise ApiException(f"invalid quantity '{quantity}'")
    return int(match.group(1))*units[match.group(2)]

def apply_pod_spec(kps, pod_spec: Dict[str, Any], predefined_pod_specs: Dict[str, Dict[str, Any]], log: logging.Logger,  kind: str) -> Dict[str, Any]:
    """applies this pod spec, as long with any predefined specs"""
    if kps:
        # apply predefined types
        if kps.type is not None:
            log.info(f"selecting predefined pod type '{kps.type}' for {kind}")
            predefined_pod_spec = predefined_pod_specs.get(kps.type)
            if predefined_pod_spec is None:
                raise StimelaCabRuntimeError(f"'{kps.type}' not found in predefined_pod_specs")
        else:
            predefined_pod_spec = {}
        # apply custom type and merge
        if predefined_pod_spec or kps.custom_pod_spec:
            pod_spec = OmegaConf.to_container(OmegaConf.merge(pod_spec, predefined_pod_spec, kps.custom_pod_spec))

        # add RAM resources
        if kps.memory is not None:
            res = pod_spec['containers'][0].setdefault('resources', {})
            # good practice to set these equal
            if kps.memory.request:
                res.setdefault('requests', {})['memory'] = kps.memory.request or kps.memory.limit
                log.info(f"setting {kind} memory request to {res['requests']['memory']}")
            if kps.memory.limit:
                res.setdefault('limits', {})['memory'] = kps.memory.limit or kps.memory.request
                log.info(f"setting {kind} memory limit to {res['limits']['memory']}")
        if kps.cpu is not None:
            res = pod_spec['containers'][0].setdefault('resources', {})
            if kps.cpu.request:
                res.setdefault('requests', {})['cpu'] = kps.cpu.request
                log.info(f"setting {kind} CPU request to {kps.cpu.request}")
            if kps.cpu.limit:
                res.setdefault('limits', {})['cpu'] = kps.cpu.limit
                log.info(f"setting {kind} CPU limit to {kps.cpu.limit}")

    return pod_spec

class StatusReporter(object):
    def __init__(self, log: logging.Logger,
                 podname: str,
                 kube: KubeBackendOptions,
                 event_handler: None,
                 update_interval: float = 1,
                 enable_metrics: bool = True):
        self.kube = kube
        self.namespace, self.kube_api, self.custom_api = get_kube_api()
        self.log = log
        self.podname = podname
        self.label_selector = f"stimela_job={podname}"
        self.event_handler = event_handler
        self.update_interval = update_interval
        self.enable_metrics = enable_metrics
        self.pod_statuses = {}
        self.reported_events = set()
        self.main_status = None
        self.pvcs = []
         # API errors added here when reported -- use this dict to avoid reissuing multiple errors
        self.api_errors_reported = {}
        self._request_timeout = (5, 5)
        self._connected = None
        self._last_connected = None
        self._last_disconnected = None

    def set_event_handler(self, event_handler):
        self.event_handler = event_handler

    def set_pod_name(self, podname):
        self.podname = podname
        self.label_selector = f"stimela_job={podname}"

    def set_pvcs(self, pvcs: Dict[str, KubeBackendOptions.Volume]):
        self.pvcs = [("PersistentVolumeClaim", pvc.name) for pvc in pvcs.values()]

    def set_main_status(self, mainstat):
        self.main_status = mainstat

    def report_api_error(self, name, exc):
        if name not in self.api_errors_reported:
            self.api_errors_reported[name] = exc
            self.log.warning(f"k8s API error for {name}: {exc}")

    @property
    def connected(self):
        return self._connected

    @connected.setter
    def connected(self, status: bool):
        if status:
            self._last_connected = datetime.now()
        else:
            self._last_disconnected = datetime.now()
        self._connected = status

    def log_events(self):
        # get list of associated pods
        try:
            pods = self.kube_api.list_namespaced_pod(namespace=self.namespace, label_selector=self.label_selector,
                                                     _request_timeout=self._request_timeout)
        except ApiException as exc:
            self.report_api_error("list_namespaced_pod", exc)
            return
        objects = [("Pod", pod.metadata.name) for pod in pods.items] + self.pvcs
        for kind, name in objects:
            # get new events
            try:
                events = self.kube_api.list_namespaced_event(namespace=self.namespace,
                                field_selector=f"involvedObject.kind={kind},involvedObject.name={name}",
                                _request_timeout=self._request_timeout)
            except ApiException as exc:
                self.report_api_error("list_namespaced_event", exc)
                return
            for event in events.items:
                if event.metadata.uid not in self.reported_events:
                    self.reported_events.add(event.metadata.uid)
                    if self.event_handler:
                        try:
                            self.event_handler(event)
                        except Exception as exc:
                            self.log.error(self.kube.debug.event_format.format(event=event))
                            raise
                    # no error from handler, report event if configured to
                    if self.kube.debug.log_events:
                        color = self.kube.debug.event_colors.get(event.type.lower()) \
                                or self.kube.debug.event_colors.get("default")
                        # escape console markup on string fields
                        for key, value in event.__dict__.items():
                            if type(value) is str:
                                setattr(event, key, escape(value))
                        self.log.info(self.kube.debug.event_format.format(event=event),
                                      extra=dict(color=color) if color else {})

    def update(self):
        from . import session_user
        metrics = []
        try:
            self.log_events()
            self.connected = True

            # update k8s stats and metrics
            pods = metrics = None
            # get pod statuses
            try:
                pods = self.kube_api.list_namespaced_pod(self.namespace,
                                                label_selector=f"stimela_user={session_user}",
                                                _request_timeout=self._request_timeout)
            except ApiException as exc:
                self.report_api_error("list_namespaced_pod", exc)
            # process statuses if we got them
            if pods:
                for pod in pods.items:
                    pname = pod.metadata.name
                    if pod.metadata.deletion_timestamp:
                        pod_status = "Terminating"
                    else:
                        pod_status = pod.status.phase
                    # if pod.status.container_statuses:
                    #     print(f"Pod: {pname}, status: {pod_status}, {[st.state for st in pod.status.container_statuses]}")
                    # get container states
                    if pod.status.container_statuses:
                        for cst in pod.status.container_statuses:
                            for state in cst.state.waiting, cst.state.terminated:
                                if hasattr(state, 'reason'):
                                    pod_status += f":{state.reason}"
                    self.pod_statuses[pname] = pod_status
            # get metrics
            if self.enable_metrics:
                try:
                    metrics = self.custom_api.list_namespaced_custom_object('metrics.k8s.io', 'v1beta1', self.namespace, 'pods',
                                                _request_timeout=self._request_timeout)
                except ApiException as exc:
                    self.report_api_error("metrics.k8s.io", exc)
        except (ConnectionError, HTTPError) as exc:
            self.connected = False
            # self.log.warning(f"disconnected: {exc}")
        # add connection status
        if not self.connected:
            interval = str(datetime.now() - self._last_connected)
            interval = interval.split(".", 1)[0]
            return [f"lost connection [red]{interval}[/red] ago"], None

        # process metrics if we got them
        if metrics:
            totals = dict(cpu=0, memory=0)
            for item in metrics['items']:
                pname = item['metadata']['name']
                if pname in self.pod_statuses:
                    for container in item['containers']:
                        usage = container['usage']
                        totals['cpu'] += resolve_unit(usage.get('cpu'), k8s_cpu_units)
                        totals['memory'] += resolve_unit(usage.get('memory'), k8s_memory_units_in_bytes)
                    # print(f"Pod: {pname}, CPU: {usage.get('cpu')}, Memory: {usage.get('memory')}")
        # add main pod/job status
        report_metrics = []
        if self.main_status:
            report_metrics.append(f"[blue]{self.main_status}[/blue]")
        elif self.podname in self.pod_statuses:
            report_metrics.append(f"[blue]{self.pod_statuses[self.podname]}[/blue]")
        # add count of running pods
        npods = len(self.pod_statuses)
        pods = ""
        nrun = sum([stat.startswith("Running") for stat in self.pod_statuses.values()])
        if nrun:
            pods += f"[green]{nrun}[/green]R"
        npend = sum([stat.startswith("Pending") for stat in self.pod_statuses.values()])
        if npend:
            pods += f"[yellow]{npend}[/yellow]P"
        nterm = sum([stat.startswith("Terminating") for stat in self.pod_statuses.values()])
        if nterm:
            pods += f"[blue]{nterm}[/blue]T"
        nsucc = sum([stat.startswith("Succeeded") for stat in self.pod_statuses.values()])
        if nsucc:
            pods += f"[green]{nsucc}[/green]S"
        nfail = sum([stat.startswith("Failed") for stat in self.pod_statuses.values()])
        if nfail:
            pods += f"[red]{nfail}[/red]F"
        nuk = npods - nrun - npend - nterm - nsucc - nfail
        if nuk:
            pods += f"[red]{nuk}[/red]U"
        if npods:
            report_metrics.append(f"pods {pods}")
        # add metrics
        if metrics:
            cores = totals['cpu']
            mem_gb = round(totals['memory'] / 2**30)
            report_metrics += [
                f"cores [green]{totals['cpu']:.2f}[/green]",
                f"mem [green]{mem_gb}[/green]G"
            ]
            stats = dict(k8s_cores=cores, k8s_mem=mem_gb)
        else:
            stats = None

        if self._last_disconnected is not None:
            report_metrics.append("reconnected")

        return report_metrics, stats





import logging
from dataclasses import dataclass
from typing import Dict, Any, Optional
from enum import Enum
from omegaconf import ListConfig, OmegaConf
from stimela.exceptions import BackendSpecificationError, BackendError
from stimela.stimelogging import log_exception
from scabha.basetypes import EmptyDictDefault, EmptyClassDefault

from .singularity import SingularityBackendOptions
from .kube import KubeBackendOptions
from .native import NativeBackendOptions
from .slurm import SlurmOptions

import stimela

## left as memo to self
# Backend = Enum("Backend", "docker singularity podman kubernetes native", module=__name__)
Backend = Enum("Backend", "singularity kube native", module=__name__)

SUPPORTED_BACKENDS = set(Backend.__members__)


def get_backend(name: str, backend_opts: Optional[Dict] = None):
    """
    Gets backend, given a name and an optional set of options for that backend.
    Returns backend module, or None if it is not available.
    """
    if name not in SUPPORTED_BACKENDS:
        return None
    backend = __import__(f"stimela.backends.{name}", fromlist=[name])
    if backend.is_available(backend_opts):
        return backend
    return None


def get_backend_status(name: str):
    if name not in SUPPORTED_BACKENDS:
        return "unknown backend"
    backend = __import__(f"stimela.backends.{name}", fromlist=[name])
    return backend.get_status()

@dataclass 
class StimelaBackendOptions(object):
    default_registry: str = "quay.io/stimela2"

    # overrides registries -- useful if you have a pull-through cache set up
    override_registries: Dict[str, str] = EmptyDictDefault()
    
    select: Any = "singularity,native"   # should be Union[str, List[str]], but OmegaConf doesn't support it, so handle in __post_init__ for now
    
    singularity: Optional[SingularityBackendOptions] = EmptyClassDefault(SingularityBackendOptions)
    kube: Optional[KubeBackendOptions] = EmptyClassDefault(KubeBackendOptions)
    native: Optional[NativeBackendOptions] = EmptyClassDefault(NativeBackendOptions)
    docker: Optional[Dict] = None  # placeholder for future impl
    slurm: Optional[SlurmOptions] = EmptyClassDefault(SlurmOptions)

    ## Resource limits applied during run -- see resource module
    rlimits: Dict[str, Any] = EmptyDictDefault()

    verbose: int = 0  # be verbose about backend selections. Higher levels mean more verbosity
    
    def __post_init__(self):
        # resolve "select" field
        if type(self.select) is str:
            if not self.select:
                self.select = []
            else:
                self.select = [x.strip() for x in self.select.split(",")]
        elif isinstance(self.select, (list, tuple, ListConfig)):
            self.select = list(self.select)
        else:
            raise BackendSpecificationError(f"invalid backend.select setting of type {self.select}")
        # provide default options for available backends
        if self.singularity is None and get_backend("singularity"):
            self.singularity = SingularityBackendOptions()
        if self.native is None and get_backend("native"):
            self.native = NativeBackendOptions()
        if self.kube is None and get_backend("kube"):
            self.kube = KubeBackendOptions()
        if self.slurm is None:
            self.slurm = SlurmOptions()

StimelaBackendSchema = OmegaConf.structured(StimelaBackendOptions)


def resolve_image_name(backend: StimelaBackendOptions, image: 'stimela.kitchen.Cab.ImageInfo'):
    """
    Resolves image name -- applies override registries, if any exist
    """
    # if image is defined, use name and registry within
    image_name = image.name
    registry_name = image.registry
    version = image.version or "latest"
    # resolve registry name
    if registry_name == "DEFAULT" or not registry_name:
        registry_name = backend.default_registry
    elif registry_name == "LOCAL":
        registry_name = ''
    # apply any registry overrides
    if registry_name in backend.override_registries:
        registry_name = backend.override_registries[registry_name]
    if registry_name:
        return f"{registry_name}/{image_name}:{version}"
    else:
        return f"{image_name}:{version}"


def _call_backends(backend_opts: StimelaBackendOptions, log: logging.Logger, method: str, desc: str, raise_exc: bool=True):
    selected = backend_opts.select or ['native']
    if type(selected) is str:
        selected = [selected]

    for engine in selected: 
        # check that backend has not been disabled
        opts = getattr(backend_opts, engine, None)
        if not opts or opts.enable:
            backend = get_backend(engine, opts)
            func = backend and getattr(backend, method, None)
            if func:
                try:
                    func(backend_opts, log)
                except BackendError as exc:
                    exc1 = BackendError(f"error {desc} {engine} backend", exc)
                    if raise_exc:
                        raise exc1 from None
                    else:
                        log_exception(exc1, log=log)

initialized = None

def init_backends(backend_opts: StimelaBackendOptions, log: logging.Logger):
    global initialized
    if initialized is None:
        initialized = backend_opts
        return _call_backends(backend_opts, log, "init", "initializing")

def close_backends(log: logging.Logger):
    global initialized
    if initialized is not None:
        result = _call_backends(initialized, log, "close", "closing")
        initialized = None
        return result

def cleanup_backends(backend_opts: StimelaBackendOptions, log: logging.Logger):
    return _call_backends(backend_opts, log, "cleanup", "cleaning up")





import os, re, subprocess, pytest
from .test_recipe import change_test_dir, run, verify_output

def test_wrangler_replace_suppress():
    print("===== expecting no errors =====")
    retcode, output = run("stimela -b native run test_wranglers.yml test_replace_suppress")
    assert retcode == 0
    print(output)
    assert verify_output(output, "Michael J. Fox", "don't need roads!")
    assert not verify_output(output, "cheetah")


def test_wrangler_force_success():
    print("===== expecting no errors =====")
    retcode, output = run("stimela -b native run test_wranglers.yml test_force_success")
    assert retcode == 0
    print(output)
    assert verify_output(output, "deliberately declared")


def test_wrangler_force_failure():
    print("===== expecting an error =====")
    retcode, output = run("stimela -b native run test_wranglers.yml test_force_failure")
    assert retcode != 0
    print(output)
    assert verify_output(output, "cab marked as failed")

    print("===== expecting an error =====")
    retcode, output = run("stimela -b native run test_wranglers.yml test_force_failure2")
    assert retcode != 0
    print(output)
    assert verify_output(output, "Nobody expected the fox!")

def test_wrangler_parse():
    print("===== expecting no errors =====")
    retcode, output = run("stimela -b native run test_wranglers.yml test_parse")
    assert retcode == 0
    print(output)
    assert verify_output(output, "The bloody cheetah ate 22 dogs!")

    print("===== expecting no errors =====")
    retcode, output = run("stimela -b native run test_wranglers.yml test_parse2")
    assert retcode == 0
    print(output)
    assert verify_output(output, "The bloody cheetah ate 22 dogs!")

    print("===== expecting no errors =====")
    retcode, output = run("stimela -b native run test_wranglers.yml test_parse3")
    assert retcode == 0
    print(output)
    assert verify_output(output, "The bloody cheetah ate 22 dogs!")




from stimela import log_exception, stimelogging
from stimela.stimelogging import log_rich_payload
from stimela.exceptions import *


def keys_from_sel_string(dictionary: Dict[str, str], sel_string: str):
    """Select keys from a dictionary based on a slice string."""

    keys = list(dictionary.keys())    

    if ':' in sel_string:
        begin, end = sel_string.split(':', 1)
        if begin:
            try:
                first = keys.index(begin)
            except ValueError as exc:
                raise StepSelectionError(f"no such step: '{begin}' (in '{sel_string}')")
        else:
            first = 0
        if end:
            try:
                last = keys.index(end)
            except ValueError as exc:
                raise StepSelectionError(f"no such step: '{end}' (in '{sel_string}')")
        else:
            last = len(keys) - 1
        selected_keys = set(keys[first: last + 1])
    else:
        if sel_string not in keys:
            raise StepSelectionError(f"no such step: '{sel_string}'")
        selected_keys = set([sel_string])

    return selected_keys



import os
from pathlib import Path
#-----------
# knicked from https://github.com/python-poetry/poetry/issues/273#issuecomment-1103812336
try:
    from importlib import metadata
except ImportError: # for Python<3.8
    import importlib_metadata as metadata
__version__ = metadata.version(__package__)
del metadata  # optional, avoids polluting the results of dir()
#------------

CONFIG = None

UID = os.getuid()
GID = os.getgid()

root = os.path.dirname(__file__)

CAB_PATH = os.path.join(root, "cargo/cab")
BASE_PATH = os.path.join(root, "cargo/base")

# Set up logging infrastructure
LOG_HOME = os.path.expanduser("~/.stimela")
# make sure directory exists
Path(LOG_HOME).mkdir(exist_ok=True)
# This is is the default log file. It logs stimela images, containers and processes
LOG_FILE = "{0:s}/stimela_logfile.json".format(LOG_HOME)

from .stimelogging import logger, log_exception



from stimela import logger, LOG_FILE, BASE, utils
from stimela.backends import docker, singularity, podman


def make_parser(subparsers):
    parser = subparsers.add_parser("kill", help='Gracefully kill stimela process(s).')

    add = parser.add_argument

    add("pid", nargs="*", help="Process ID")

    parser.set_defaults(func=kill)


def kill(args, conf):
    pass



import os, os.path, re, logging, copy, shutil, time
from typing import Any, Tuple, List, Dict, Optional, Union
from dataclasses import dataclass
from omegaconf import MISSING, OmegaConf, DictConfig, ListConfig
from omegaconf.errors import OmegaConfBaseException
from collections import OrderedDict
from contextlib import nullcontext
from rich.markup import escape

from stimela.config import EmptyDictDefault, EmptyListDefault
import stimela
from stimela import log_exception, stimelogging, task_stats
from stimela.stimelogging import log_rich_payload
from stimela.backends import StimelaBackendSchema, runner
from stimela.exceptions import *
import scabha.exceptions
from scabha.exceptions import SubstitutionError, SubstitutionErrorList
from scabha.validate import evaluate_and_substitute, evaluate_and_substitute_object, Unresolved, join_quote
from scabha.substitutions import SubstitutionNS, substitutions_from 
from scabha.basetypes import UNSET, Placeholder, MS, File, Directory, SkippedOutput
from .cab import Cab, get_cab_schema

Conditional = Optional[str]


def resolve_dotted_reference(key, base, current, context): 
    """helper function to look up a key like a.b.c in a nested dict-like structure"""
    path = key.split('.')
    if path[0]:
        section = base
    else:
        if not current:
            raise NameError(f"{context}: leading '.' not permitted here")
        section = current
        path = path[1:]
        if not path:
            raise NameError(f"{context}: '.' not permitted")
    varname = path[-1]
    for element in path[:-1]:
        if not element:
            raise NameError(f"{context}: '..' not permitted")
        if element in section:
            section = section[element]
        else:
            raise NameError(f"{context}: '{element}' in '{key}' is not a valid config section")
    return section, varname

OUTPUTS_EXISTS = "exist"
OUTPUTS_FRESH = "fresh"

@dataclass
class Step:
    """Represents one processing step of a recipe"""
    cab: Optional[Any] = None                       # if not None, this step is a cab and this is the cab name
    recipe: Optional[Any] = None                    # if not None, this step is a nested recipe
    params: Dict[str, Any] = EmptyDictDefault()     # assigns parameter values
    info: Optional[str] = None                      # comment or info string
    skip: Optional[str] = None                      # if this evaluates to True, step is skipped.  
    skip_if_outputs: Optional[str] = None           # skip if outputs "exist' or "fresh"
    tags: List[str] = EmptyListDefault()

    name: str = ''                                  # step's internal name
    fqname: str = ''                                # fully-qualified name e.g. recipe_name.step_label

    assign: Dict[str, Any] = EmptyDictDefault()     # assigns recipe-level variables when step is executed

    assign_based_on: Dict[str, Any] = EmptyDictDefault()
                                                    # assigns recipe-level variables when step is executed based on value of another variable

    # optional backend settings
    backend: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        self.fqname = self.fqname or self.name
        if not bool(self.cab) and not bool(self.recipe):
            raise StepValidationError(f"step '{self.name}': step must specify either a cab or a nested recipe")
        if bool(self.cab) == bool(self.recipe):
            raise StepValidationError(f"step '{self.name}': step can't specify both a cab and a nested recipe")
        self.cargo = self.config = None
        self.tags = set(self.tags)
        # check backend setting
        if self.backend:
            try:
                OmegaConf.merge(StimelaBackendSchema, self.backend)
            except OmegaConfBaseException as exc:
                raise StepValidationError(f"step '{self.name}': invalid backend setting", exc)
        # convert params into standard dict, else lousy stuff happens when we insert non-standard objects
        if isinstance(self.params, DictConfig):
            self.params = OmegaConf.to_container(self.params)
        # after (pre)validation, this contains parameter values
        self.validated_params = None
        # parameters protected from assignment (because they've been set on the command line, presumably)
        self._assignment_overrides = set()
        if self.skip_if_outputs and self.skip_if_outputs not in (OUTPUTS_EXISTS, OUTPUTS_FRESH):
            raise StepValidationError(f"step '{self.name}': invalid 'skip_if_outputs={self.skip_if_outputs}' setting")
        # the "skip" attribute is reevaluated at runtime since it may contain substitutions, but if it's set to a bool
        # constant, self._skip will be preset already
        # validate skip attribute
        if self.skip in {"True", "true", "1"}:
            self._skip = True
        elif self.skip in {"False", "false", "0", "", None}:
            self._skip = False
        else:
            # otherwise, self._skip stays at None, and will be re-evaluated at runtime
            self._skip = None
        
    def summary(self, params=None, recursive=True, ignore_missing=False, inputs=True, outputs=True):
        summary_params = OrderedDict()
        for name, value in (params or self.validated_params or self.params).items():
            schema = self.cargo.inputs_outputs[name]
            if (inputs and (schema.is_input or schema.is_named_output)) or \
                (outputs and schema.is_output):
                summary_params[name] = value
        return self.cargo and self.cargo.summary(recursive=recursive, params=summary_params, ignore_missing=ignore_missing)

    @property
    def finalized(self):
        return self.cargo is not None

    @property
    def missing_params(self):
        return OrderedDict([(name, schema) for name, schema in self.cargo.inputs_outputs.items() 
                            if schema.required and name not in self.validated_params])

    @property
    def invalid_params(self):
        return [name for name, value in self.validated_params.items() if isinstance(value, scabha.exceptions.Error)]

    @property
    def unresolved_params(self):
        return [name for name, value in self.validated_params.items() if isinstance(value, Unresolved) and not isinstance(value, Placeholder)]

    @property
    def inputs(self):
        return self.cargo.inputs

    @property
    def outputs(self):
        return self.cargo.outputs

    @property
    def inputs_outputs(self):
        return self.cargo.inputs_outputs

    @property
    def log(self):
        """Logger object passed from cargo"""
        return self.cargo and self.cargo.log
    
    @property
    def logopts(self):
        """Logger options passed from cargo"""
        return self.cargo and self.cargo.logopts

    @property
    def nesting(self):
        """Logger object passed from cargo"""
        return self.cargo and self.cargo.nesting

    def update_parameter(self, name, value):
        self.params[name] = value

    def unset_parameter(self, name):
        if name in self.params:
            del self.params[name]

    def update_log_options(self, **options):
        from .recipe import Recipe
        for setting, value in options.items():
            try:
                self.logopts[setting] = value
            except Exception as exc:
                raise AssignmentError(f"invalid {self.fqname}.log.{setting} setting", exc)
        # propagate to children
        if isinstance(self.cargo, Recipe):
            self.cargo.update_log_options(**options)

    _instantiated_cabs = {}

    def finalize(self, config=None, log=None, fqname=None, backend=None, nesting=0):
        from .recipe import Recipe, RecipeSchema
        if not self.finalized:
            if fqname is not None:
                self.fqname = fqname
            self.config = config = config or stimela.CONFIG

            # if recipe, validate the recipe with our parameters
            if self.recipe:
                # first, if it is a string, look it up in library
                recipe_name = f"{self.fqname}:recipe"
                if type(self.recipe) is str:
                    recipe_name = f"nested recipe '{self.recipe}'"
                    # undotted name -- look in lib.recipes
                    if '.' not in self.recipe:
                        if self.recipe not in self.config.lib.recipes:
                            raise StepValidationError(f"recipe '{self.recipe}' not found in lib.recipes")
                        self.recipe = self.config.lib.recipes[self.recipe]
                    # dotted name -- look in config
                    else: 
                        section, var = resolve_dotted_reference(self.recipe, config, current=None, context=f"step '{self.name}'")
                        if var not in section:
                            raise StepValidationError(f"recipe '{self.recipe}' not found")
                        self.recipe = section[var]
                    # self.recipe is now hopefully a DictConfig or a Recipe object, so fall through below to validate it 
                # instantiate from omegaconf object, if needed
                if type(self.recipe) is DictConfig:
                    try:
                        self.recipe = Recipe(**OmegaConf.unsafe_merge(RecipeSchema.copy(), self.recipe))
                    except OmegaConfBaseException as exc:
                        raise StepValidationError(f"error in recipe '{recipe_name}", exc)
                elif not isinstance(self.recipe, Recipe):
                    raise StepValidationError(f"recipe field must be a string or a nested recipe, got {type(self.recipe)}")
                self.cargo = self.recipe
            else:
                if type(self.cab) is str:
                    if self.cab in self._instantiated_cabs:
                        self.cargo = copy.copy(self._instantiated_cabs[self.cab])
                    else:
                        if self.cab not in self.config.cabs:
                            raise StepValidationError(f"unknown cab '{self.cab}'")
                        try:
                            self._instantiated_cabs[self.cab] = Cab(**config.cabs[self.cab])
                            self.cargo = copy.copy(self._instantiated_cabs[self.cab])
                        except Exception as exc:
                            raise StepValidationError(f"error in cab '{self.cab}'", exc)
                else:
                    if type(self.cab) is DictConfig:
                        cab_name = f"{self.fqname}:cab"
                        try:
                            self.cab = Cab(**OmegaConf.unsafe_merge(get_cab_schema().copy(), self.cab))
                        except OmegaConfBaseException as exc:
                            raise StepValidationError(f"error in cab '{cab_name}", exc)
                    elif not isinstance(self.cab, Cab):
                        raise StepValidationError(f"cab field must be a string or an inline cab, got {type(self.cab)}")
                    self.cargo = self.cab
            self.cargo.name = self.cargo.name or self.name

            # flatten parameters
            self.cargo.apply_dynamic_schemas(self.params)
            self.params = self.cargo.flatten_param_dict(OrderedDict(), self.params)

            # if logger is not provided, then init one
            if log is None:
                log = stimela.logger().getChild(self.fqname)
                log.propagate = False

            # finalize the cargo
            self.cargo.finalize(config, log=log, fqname=self.fqname, backend=backend, nesting=nesting)

            # build dictionary of defaults from cargo
            self.defaults = {name: schema.default for name, schema in self.cargo.inputs_outputs.items() 
                             if schema.default is not UNSET and not isinstance(schema.default, Unresolved) }
            self.defaults.update(**self.cargo.defaults)
            
            # set missing parameters from defaults
            for name, value in self.defaults.items():
                if name not in self.params:
                    self.params[name] = value

            # check for valid backend
            backend_opts = OmegaConf.to_object(OmegaConf.merge(
                StimelaBackendSchema,
                backend or {}, 
                self.cargo.backend or {}, 
                self.backend or {}))
            runner.validate_backend_settings(backend_opts, log=log)

    def prevalidate(self, subst: Optional[SubstitutionNS]=None, root=False, backend=None):
        self.finalize(backend=backend)
        # apply dynamic schemas
        params = self.params
        if self.cargo.has_dynamic_schemas:
            # prevalidate in order to resolve substitutions in existing parameters
            params = self.cargo.prevalidate(params, subst, root=root)
            self.cargo.apply_dynamic_schemas(params, subst)
            # will prevvalidate again below based on these updated schemas
        # validate cab or recipe
        params = self.validated_params = self.cargo.prevalidate(params, subst, root=root)
        # add missing outputs
        for name, schema in self.cargo.outputs.items():
            if name not in params and schema.required:
                params[name] = UNSET(name)
        self.log.debug(f"{self.cargo.name}: {len(self.missing_params)} missing, "
                        f"{len(self.invalid_params)} invalid and "
                        f"{len(self.unresolved_params)} unresolved parameters")
        if self.invalid_params:
            raise StepValidationError(f"step '{self.name}': {self.cargo.name} has the following invalid parameters: {join_quote(self.invalid_params)}")
        return params

    def log_summary(self, level, title, color=None, ignore_missing=True, inputs=False, outputs=False):
        extra = dict(color=color)
        if self.log.isEnabledFor(level):
            self.log.log(level, f"### {title}", extra=extra)
            for line in self.summary(recursive=False, inputs=inputs, outputs=outputs, ignore_missing=ignore_missing):
                self.log.log(level, escape(line), extra=extra)

    def log_exception(self, exc, severity="error", log=None):
        log_exception(exc, severity=severity, log=log or self.log)

    def assign_value(self, key: str, value: Any, override: bool = False):
        """assigns parameter value or nested variable value to this step

        Args:
            key (str): name
            value (Any): value
            override (bool): If True, value will override all future assignments (used for command-line overrides)
                             Defaults to False.
        """
        # ignore assignment if an override assignment was done earlier
        if key in self._assignment_overrides and not override:
            return
        if override:
            self._assignment_overrides.add(key)
        # assigning parameter directly? Add to self.params
        if key in self.inputs_outputs:
            self.params[key] = value
            # and remove from prevalidated params
            if self.validated_params and key in self.validated_params:
                del self.validated_params[key]
        # else delegate to cargo to assign
        else:
            try:
                self.cargo.assign_value(key, value, override=override)
            except ScabhaBaseException as exc:
                raise AssignmentError(f"{self.name}: invalid assignment {key}={value}", exc)


    def build(self, backend=None, rebuild=False, build_skips=False, log: Optional[logging.Logger] = None):
        # skipping step? ignore the build
        if self.skip is True and not build_skips:
            return
        backend = OmegaConf.merge(backend or {}, self.cargo.backend or {}, self.backend or {})
        log = log or self.log
        # recurse into sub-recipe 
        from .recipe import Recipe
        if type(self.cargo) is Recipe:
            return self.cargo.build(backend, rebuild=rebuild, build_skips=build_skips, log=log)
        # else build 
        else:
            # validate backend settings and call the build function
            try:
                backend_opts = OmegaConf.merge(self.config.opts.backend, backend)
                if getattr(backend_opts, 'verbose', 0):
                    opts_yaml = OmegaConf.to_yaml(backend_opts)
                    log_rich_payload(self.log, "effective backend settings are", opts_yaml, syntax="yaml") 
                backend_opts = OmegaConf.to_object(OmegaConf.merge(stimela.CONFIG.opts.backend, backend_opts))
                backend_runner = runner.validate_backend_settings(backend_opts, log=log)
            except Exception as exc:
                newexc = BackendError("error validating backend settings", exc)
                raise newexc from None
            log.info(f"building image for step '{self.fqname}' using the {backend_runner.backend_name} backend")
            with task_stats.declare_subtask(self.name):
                return backend_runner.build(self.cargo, log=log, rebuild=rebuild)


    def run(self, backend: Optional[Dict] = None, subst: Optional[Dict[str, Any]] = None, 
            is_outer_step: bool=False,
            parent_log: Optional[logging.Logger] = None) -> Dict[str, Any]:
        """executes the step

        Args:
            backend (Dict, optional): Backend settings inherited from parent.
            subst (Dict[str, Any], optional): Substitution namespace. Defaults to None.
            parent_log (logging.Logger, optional): parent logger for parent-related messages. Defaults to using the step logger if not supplied.

        Raises:
            StepValidationError: _description_
            StepValidationError: _description_
            StepValidationError: _description_
            StimelaCabRuntimeError: _description_
            RuntimeError: _description_
            StepValidationError: _description_

        Returns:
            Dict[str, Any]: step outputs
        """

        from .recipe import Recipe

        # some messages go to the parent logger -- if not defined, default to our own logger
        if parent_log is None:
            parent_log = self.log

        backend = OmegaConf.merge(backend or {}, self.cargo.backend or {}, self.backend or {})

        # validate backend settings
        try:
            backend_opts = OmegaConf.merge(self.config.opts.backend, backend)
            backend_opts = evaluate_and_substitute_object(backend_opts, subst, 
                                                          recursion_level=-1, location=[self.fqname, "backend"])
            if not is_outer_step and backend_opts.verbose:
                opts_yaml = OmegaConf.to_yaml(backend_opts)
                log_rich_payload(self.log, "current backend settings are", opts_yaml, syntax="yaml") 
            backend_opts = OmegaConf.merge(StimelaBackendSchema, backend_opts)
            backend_opts = OmegaConf.to_object(backend_opts)
            backend_runner = runner.validate_backend_settings(backend_opts, log=self.log)
        except Exception as exc:
            newexc = BackendError("error validating backend settings", exc)
            raise newexc from None

        # if step is being explicitly skipped, omit from profiling, and drop info/warning messages to debug level
        explicit_skip = self.skip is True 
        if explicit_skip:
            context = nullcontext()
            parent_log_info = parent_log_warning = parent_log.debug
        else:
            context = task_stats.declare_subtask(self.name, hide_local_metrics=backend_runner.is_remote)
            stimelogging.declare_chapter(f"{self.fqname}")
            parent_log_info, parent_log_warning = parent_log.info, parent_log.warning

        if self.validated_params is None:
            self.prevalidate(self.params)

        with context:
            # evaluate the skip attribute (it can be a formula and/or a {}-substititon)
            skip = self._skip
            if self._skip is None and subst is not None:
                skip = evaluate_and_substitute_object(self.skip, subst, 
                                                      location=[self.fqname, "skip"])
                if skip is UNSET:  # skip: =IFSET(recipe.foo) will return UNSET
                    skip = False
                self.log.debug(f"dynamic skip attribute evaluation returns {skip}")
                # formulas with unset variables return UNSET instance
                if isinstance(skip, UNSET):
                    if skip.errors:
                        raise StepValidationError(f"{self.fqname}.skip: error evaluating '{self.skip}'", skip.errors)
                    else:
                        raise StepValidationError(f"{self.fqname}.skip: error evaluating '{self.skip}'", SubstitutionError(f"unknown variable '{skip.value}'"))

            # Since prevalidation will have populated default values for potentially missing parameters, use those values
            # For parameters that aren't missing, use whatever value that was suplied
            # preserve order of specified params, to allow ordered substitutions to occur
            params = self.params.copy()
            params.update([(key, value) for key, value in self.validated_params.items() if key not in params])

            skip_warned = False   # becomes True when warnings are given

            self.log.debug(f"validating inputs {subst and list(subst.keys())}")
            validated = None
            try:
                params = self.cargo.validate_inputs(params, loosely=skip, remote_fs=backend_runner.is_remote_fs, subst=subst)
                validated = True

            except ScabhaBaseException as exc:
                severity = "warning" if skip else "error"
                level = logging.WARNING if skip else logging.ERROR
                if not exc.logged and not explicit_skip:
                    if type(exc) is SubstitutionErrorList:
                        self.log_exception(StepValidationError(f"unresolved {{}}-substitution(s) in inputs:", exc.nested), 
                                           severity=severity)
                        # for err in exc.errors:
                        #     self.log.log(level, f"  {err}")
                    else:
                        self.log_exception(StepValidationError(f"error validating inputs:", exc), 
                                           severity=severity)
                    exc.logged = True
                if not explicit_skip:
                    self.log_summary(level, "summary of inputs follows", color="WARNING", inputs=True)
                # raise up, unless step is being skipped
                if skip:
                    parent_log_warning("since the step is being skipped, this is not fatal")
                    skip_warned = True
                else:
                    raise

            self.validated_params.update(**params)

            # log inputs
            if validated and not skip:
                self.log_summary(logging.INFO, "validated inputs", color="GREEN", ignore_missing=True, inputs=True)
                if subst is not None:
                    subst.current = params

            ## check for (a) invalid params (b) unresolved inputs 
            # (c) unresolved outputs of File/MS/Directory type 
            invalid = self.invalid_params
            for name in self.unresolved_params:
                schema = self.cargo.inputs_outputs[name]
                if schema.is_input or schema.is_named_output:
                    invalid.append(name)
            if invalid:
                if skip:
                    parent_log_warning(f"invalid inputs: {join_quote(invalid)}")
                    if not skip_warned:
                        parent_log_warning("since the step was skipped, this is not fatal")
                        skip_warned = True
                else:
                    raise StepValidationError(f"step '{self.name}': invalid inputs: {join_quote(invalid)}", log=self.log)

            ## check if we need to skip based on existing/fresh file outputs
            skip_if_outputs = self.skip_if_outputs
            # don't check if skipping anyway
            if skip:
                skip_if_outputs = None
            # don't check if remote filesystem
            elif backend_runner.is_remote_fs:
                parent_log_info(f"ignoring skip_if_outputs: {skip_if_outputs} because backend has remote filesystem")
                skip_if_outputs = None
            # don't check if force-disabled
            elif (skip_if_outputs == OUTPUTS_EXISTS and stimela.CONFIG.opts.disable_skips.exist) or \
                    (skip_if_outputs == OUTPUTS_FRESH and stimela.CONFIG.opts.disable_skips.fresh):
                parent_log_info(f"ignoring skip_if_outputs: {skip_if_outputs} because it has been force-disabled")
                skip_if_outputs = None

            ## if skip on fresh outputs is in effect, find mtime of most recent input 
            if skip_if_outputs:
                # max_mtime will remain 0 if we're not echecking for freshness, or if there are no file-type inputs
                max_mtime, max_mtime_path = 0, None
                if skip_if_outputs == OUTPUTS_FRESH:
                    parent_log_info("checking if file-type outputs of step are fresh")
                    for name, value in params.items():
                        schema = self.inputs_outputs[name]
                        if schema.is_input and not schema.skip_freshness_checks:
                            if schema.is_file_type:
                                values = [value]
                            elif schema.is_file_list_type:
                                values = value
                            else:
                                continue
                            for filename in values:
                                if type(filename) is str and os.path.exists(filename):
                                    mtime = os.path.getmtime(filename)
                                    if mtime > max_mtime:
                                        max_mtime = mtime
                                        max_mtime_path = filename
                    if max_mtime:
                        parent_log_info(f"  most recently modified input is {max_mtime_path} ({time.ctime(max_mtime)})")
                else:
                    parent_log_info("checking if file-type outputs of step exist")

                ## now go through outputs -- all_exist flag will be cleared if we find one that doesn't exist,
                ## or is older than an input
                all_exist = True
                for name, schema in self.outputs.items():
                    # ignore outputs not in params (implicit outputs will be already in there thanks to validation above)
                    if name in params:
                        # check for files or lists of files, and skip otherwise
                        if schema.is_file_type:
                            filenames = [params[name]]
                        elif schema.is_file_list_type:
                            filenames = params[name]
                            # empty list of files treated as non-existing output
                            if not filenames:
                                if schema.must_exist:
                                    all_exist = False
                                    parent_log_info(f"  {name}: no existing file(s)")
                                    break  # abort the check
                                else:
                                    parent_log_info(f"  {name}: no existing file(s), but they are not required")
                                    continue
                        else:
                            continue # go on to next parameter
                        # collect messages rather than logging them directly, to avoid log diarrhea for long file lists
                        messages = []
                        # ok, we have a list of files to check
                        for num, value in enumerate(filenames):
                            if type(value) is not str:  # skip funny values that aren't strings
                                continue
                            # form up label for messages
                            label = f"{name}[{num}]" if schema.is_file_list_type else name
                            if os.path.exists(value):
                                # max_mtime==0 means we're only checking for existence, not freshness
                                if max_mtime:
                                    if schema.skip_freshness_checks:
                                        messages.append(f"{label} = {value} marked as skipped from freshness checks")
                                    else:
                                        mtime = os.path.getmtime(value)
                                        if mtime < max_mtime:
                                            parent_log_info(f"{label} = {value} is not fresh")
                                            all_exist = False
                                            break
                                        else:
                                            messages.append(f"{label} = {value} is fresh")
                                else:
                                    messages.append(f"{label} = {value} exists")
                            elif schema.must_exist is not False:
                                all_exist = False
                                parent_log_info(f"  {label} = {value} doesn't exist")
                                break
                            else:
                                messages.append(f"{label} = {value} doesn't exist, but is not required to")
                        # abort the checks if we encountered a fail
                        if not all_exist:
                            break
                        # else log the collected messages
                        if len(messages) > 2:
                            messages = [messages[0], "  ...", messages[-1]]
                        for msg in messages:
                            parent_log_info(f"  {msg}")
                if all_exist:
                    parent_log_info("all required outputs are OK, skipping this step")
                    skip = True

            if not skip:
                # check for outputs that need removal
                if not backend_runner.is_remote_fs:
                    for name, schema in self.outputs.items():
                        if name in params and schema.path_policies.remove_if_exists and schema.is_file_type:
                            path = params[name]
                            if type(path) is str and os.path.exists(path):
                                if os.path.isdir(path) and not os.path.islink(path):
                                    shutil.rmtree(path)
                                else:
                                    os.unlink(path)

                if type(self.cargo) is Recipe:
                    self.cargo._run(params, subst, backend=backend)
                elif type(self.cargo) is Cab:
                    cabstat = backend_runner.run(self.cargo, params=params, log=self.log, subst=subst, fqname=self.fqname)
                    # check for runstate
                    if cabstat.success is False:
                        raise StimelaCabRuntimeError(f"error running cab '{self.cargo.name}'", cabstat.errors)
                    for msg in cabstat.warnings:
                        self.log.warning(f"cab '{self.cargo.name}': {msg}")
                    params.update(**cabstat.outputs)
                else:
                    raise RuntimeError("step '{self.name}': unknown cargo type")
            else:
                if self._skip is None and subst is not None:
                    parent_log_info(f"skipping step based on conditonal settings")
                else:
                    parent_log.debug("skipping step based on explicit setting")

            self.log.debug(f"validating outputs")
            validated = False

            try:
                params = self.cargo.validate_outputs(params, loosely=skip,remote_fs=backend_runner.is_remote_fs, subst=subst)
                validated = True
            except ScabhaBaseException as exc:
                severity = "warning" if skip else "error"
                level = logging.WARNING if self.skip else logging.ERROR
                if not exc.logged:
                    if type(exc) is SubstitutionErrorList:
                        self.log_exception(StepValidationError(f"unresolved {{}}-substitution(s) in outputs:", exc.nested), severity=severity)
                        # for err in exc.errors:
                        #     self.log.log(level, f"  {err}")
                    else:
                        self.log_exception(StepValidationError(f"error validating outputs:", exc), severity=severity)
                    exc.logged = True
                # raise up, unless step is being skipped
                if skip:
                    self.log.warning("since the step was skipped, this is not fatal")
                else:
                    self.log_summary(level, "failed outputs", color="WARNING", inputs=False, outputs=True)
                    raise

            if validated:
                self.validated_params.update(**params)
                if subst is not None:
                    subst.current._merge_(params)
                self.log_summary(logging.DEBUG, "validated outputs", ignore_missing=True, outputs=True)

            # bomb out if an output was invalid
            invalid = [name for name in self.invalid_params + self.unresolved_params 
                        if name in self.cargo.outputs and self.cargo.outputs[name].required is not False]
            if invalid:
                if skip:
                    parent_log_warning(f"invalid outputs: {join_quote(invalid)}")
                    parent_log_warning("since the step was skipped, this is not treated as an error for now, but may cause errors downstream")
                    for key in invalid:
                        params[key] = SkippedOutput(key)
                else:
                    # check if invalid steps are due to subrecipe with skipped steps, ignpre those
                    truly_invalid = [name for name in invalid if not isinstance(params.get(name), SkippedOutput)]
                    if truly_invalid:
                        raise StepValidationError(f"invalid outputs: {join_quote(truly_invalid)}", log=self.log)
                    parent_log_warning(f"invalid outputs: {join_quote(invalid)}")
                    parent_log_warning("since some sub-steps were skipped, this is not treated as an error for now, but may cause errors downstream")

        return params




import importlib
import os, os.path, time, platform, traceback
from typing import Any, List, Dict, Optional
from dataclasses import dataclass, field
from omegaconf.omegaconf import OmegaConf
from omegaconf.errors import OmegaConfBaseException
from collections import OrderedDict
import psutil

from yaml.error import YAMLError
import stimela
from stimela.exceptions import *
from stimela import log_exception

from scabha import configuratt
from scabha.basetypes import EmptyDictDefault, EmptyListDefault, EmptyClassDefault
from stimela.backends import StimelaBackendOptions

@dataclass 
class StimelaLogConfig(object):
    enable: bool = True                          
    name: str = "log-{info.fqname}"          # Default name for log file. info dict and {config.x.y} is substituted.
    ext: str = ".txt"                        # Default extension for log file.
    dir: str = "."                               # Default directory for log files

    symlink: Optional[str] = None                # Will make named symlink to the log directory. A useful pattern is e.g. dir="logs-{config.run.datetime}", symlink="logs",
                                            # then each run has its own log dir, and "logs" always points to the latest one

    # how deep to nest individual log files. 0 means one log per recipe, 1 means one per step, 2 per each substep, etc. 
    nest: int = 999                             
    
    level: str = "INFO"                          # level at which we log
    

## overall Stimela config schema

import stimela.backends


@dataclass
class StimelaProfilingOptions(object):
    print_depth: int = 9999
    unroll_loops: bool = False

@dataclass
class StimelaDisableSkipOptions(object):
    fresh: bool = False
    exist: bool = False

@dataclass
class StimelaOptions(object):
    backend: StimelaBackendOptions = EmptyClassDefault(StimelaBackendOptions)
    log: StimelaLogConfig = EmptyClassDefault(StimelaLogConfig)
    ## list of paths to search with _include
    include: List[str] = EmptyListDefault()
    ## Miscellaneous runtime options (runtime.casa, etc.)     
    runtime: Dict[str, Any] = EmptyDictDefault()    
    ## Profiling options
    profile: StimelaProfilingOptions = EmptyClassDefault(StimelaProfilingOptions)
    ## Disables skip_if_outputs checks
    disable_skips: StimelaDisableSkipOptions = EmptyClassDefault(StimelaDisableSkipOptions)

def DefaultDirs():
    return field(default_factory=lambda:dict(indir='.', outdir='.'))

_CONFIG_BASENAME = "stimela.conf"
_STIMELA_CONFDIR = os.path.os.path.expanduser("~/.stimela")

# check for cultcargo module
try:
    ccmod = importlib.import_module("cultcargo")
except ImportError:
    ccmod = None

# dict of config file locations to check, in order of preference
CONFIG_LOCATIONS = OrderedDict(
    package = os.path.join(os.path.dirname(__file__), _CONFIG_BASENAME),
    local   = _CONFIG_BASENAME,
    venv    = os.environ.get('VIRTUAL_ENV', None) and os.path.join(os.environ['VIRTUAL_ENV'], _CONFIG_BASENAME),
    stimela = os.path.isdir(_STIMELA_CONFDIR) and os.path.join(_STIMELA_CONFDIR, _CONFIG_BASENAME),
    cultcargo = ccmod and os.path.join(os.path.dirname(ccmod.__file__), _CONFIG_BASENAME),
    user    = os.path.join(os.path.expanduser("~/.config"), _CONFIG_BASENAME),
)

# set to the config file that was actually found
CONFIG_LOADED = None

# set to the set of config dependencies
CONFIG_DEPS = None

# stimela base directory
STIMELA_DIR = os.path.dirname(stimela.__file__)


def merge_extra_config(conf, newconf):
    from stimela import logger

    if 'cabs' in newconf:
        for cab in newconf.cabs:
            if cab in conf.cabs:
                logger().warning(f"changing definition of cab '{cab}'")
    return OmegaConf.unsafe_merge(conf, newconf)


StimelaConfigSchema = None

ConfigExceptionTypes = (configuratt.ConfigurattError, OmegaConfBaseException, YAMLError)

def get_initial_deps():
    dependencies = configuratt.ConfigDependencies()

    # add ourselves to dependencies
    dependencies.add(STIMELA_DIR, version=stimela.__version__)              
    dependencies.add(__file__, origin=STIMELA_DIR)              
    dependencies.add(configuratt.__file__, origin=STIMELA_DIR)  

    return dependencies


def load_config(extra_configs: List[str], extra_dotlist: List[str] = [], include_paths: List[str] = [],
                verbose: bool = False, use_sys_config: bool = True):

    # # disable OmegaConf resolvers
    # for name in "oc.create", "oc.decode", "oc.deprecated", "oc.env", "oc.select", "oc.dict.keys", "oc.dict.values":
    #     print(f"clearing {name}")
    #     print(OmegaConf.clear_resolver(name))

    log = stimela.logger()

    configuratt.common.PACKAGE_VERSION = f"stimela=={stimela.__version__}"
    # set up include paths

    # stadard system paths
    configuratt.PATH[0:0] = [os.path.expanduser("~/lib/stimela"), "/usr/lib/stimela", "/usr/local/lib/stimela"]
    if 'VIRTUAL_ENV' in os.environ:
        configuratt.PATH.insert(0, os.environ['VIRTUAL_ENV'])
        configuratt.PATH.insert(0, os.path.join(os.environ['VIRTUAL_ENV'], "lib/stimela"))
    if os.path.isdir(_STIMELA_CONFDIR):
        configuratt.PATH.insert(0, _STIMELA_CONFDIR)

    # add paths from command line and environment variable
    paths = [os.path.expanduser(path) for path in include_paths]
    envpaths = os.environ.get("STIMELA_INCLUDE")
    if envpaths:
        paths += envpaths.split(':')

    if paths:
        log.info(f"added include paths: {' '.join(paths)}")
        configuratt.PATH[0:0] = paths

    if verbose:
        log.info(f"include paths are {':'.join(configuratt.PATH)}")

    extra_cache_keys = list(extra_dotlist) + configuratt.PATH

    STIMELA_DIR = os.path.dirname(stimela.__file__)
    from stimela.kitchen.cab import Cab, ImageInfo

    global StimelaConfigSchema, StimelaLibrary, StimelaConfig
    @dataclass
    class StimelaLibrary(object):
        params: Dict[str, Any] = EmptyDictDefault()
        recipes: Dict[str, Any] = EmptyDictDefault()
        steps: Dict[str, Any] = EmptyDictDefault()
        misc: Dict[str, Any] = EmptyDictDefault()
        wisdom: Dict[str, Any] = EmptyDictDefault()
        

    @dataclass 
    class StimelaConfig:
        images: Dict[str, ImageInfo] = EmptyDictDefault()
        lib: StimelaLibrary = EmptyClassDefault(StimelaLibrary)
        cabs: Dict[str, Cab] = EmptyDictDefault()
        opts: StimelaOptions = EmptyClassDefault(StimelaOptions)
        vars: Dict[str, Any] = EmptyDictDefault()
        run:  Dict[str, Any] = EmptyDictDefault()
        

    base_configs = lib_configs = cab_configs = []

    if use_sys_config:
        sys_configs = [config_file for config_file in CONFIG_LOCATIONS.values()
                        if config_file and os.path.exists(config_file)]
    else:
        sys_configs = []

    all_configs = base_configs + lib_configs + cab_configs + sys_configs + list(extra_configs)

    conf, dependencies = configuratt.load_cache(all_configs, extra_keys=extra_cache_keys, verbose=verbose) 

    if conf is not None:
        log.info("loaded full configuration from cache")
    else:
        log.info("loading configuration")
        dependencies = get_initial_deps()

        # start with empty structured config containing schema
        cab_schema = OmegaConf.structured(Cab)
        opts_schema = OmegaConf.structured(StimelaOptions)

        StimelaConfigSchema = OmegaConf.structured(StimelaConfig)

        conf = StimelaConfigSchema.copy()

        # merge lib/params/*yaml files into the config
        try:
            conf.lib.params, deps = configuratt.load_nested(lib_configs,  use_sources=[conf], 
                                                                location='lib.params', include_path='_path') 
            dependencies.update(deps)
        except Exception as exc:
            if verbose:
                traceback.print_exc()
            log_exception(ConfigError("error loading lib.params configuration", exc))
            return None

        # merge all cab/*/*yaml files into the config, under cab.taskname
        try:
            conf.cabs, deps = configuratt.load_nested(cab_configs, use_sources=[conf], structured=cab_schema, 
                                                        nameattr='name', include_path='_path', location='cabs', 
                                                        use_cache=False, verbose=verbose)
            dependencies.update(deps)
        except Exception as exc:
            if verbose:
                traceback.print_exc()
            log_exception(ConfigError("error loading cabs configuration", exc))
            return None

        conf.opts = opts_schema

        def _load(conf, config_file):
            global CONFIG_LOADED
            log.info(f"loading config from {config_file}")
            try:
                newconf, deps = configuratt.load(config_file, use_sources=[conf], verbose=verbose, use_cache=False)
                dependencies.update(deps)
                conf = merge_extra_config(conf, newconf)
                if not CONFIG_LOADED:
                    CONFIG_LOADED = config_file
            except ConfigExceptionTypes as exc:
                if verbose:
                    traceback.print_exc()
                log_exception(ConfigError("error reading {config_file}", exc))
            return conf

        # add standard configs 
        for config_file in sys_configs:
            conf = _load(conf, config_file)

        # add local configs
        for path in extra_configs:
            conf = _load(conf, path)

        if not CONFIG_LOADED:
            log.info("no user-supplied configuration files given, using defaults")

        # dependencies.replace((base_configs_glob, cab_configs_glob, lib_configs_glob), STIMELA_DIR)

        configuratt.save_cache(all_configs, conf, dependencies, extra_keys=extra_cache_keys, verbose=verbose)

    # add dotlist settings
    if extra_dotlist:
        try:
            dotlist_conf = OmegaConf.from_dotlist(extra_dotlist)
            conf = OmegaConf.unsafe_merge(conf, dotlist_conf)
        except Exception as exc:
            if verbose:
                traceback.print_exc()
            log_exception(f"error applying command-line config settings", exc)
            return None

    # add runtime info
    _ds = time.strftime("%Y%m%d")
    _ts = time.strftime("%H%M%S")
    runtime = dict(
        date=_ds, 
        time=_ts, datetime=f"{_ds}-{_ts}", 
        ncpu=psutil.cpu_count(logical=True),
        node=platform.node().split('.', 1)[0],
        hostname=platform.node(), 
        env={key: value.replace('${', r'\${') for key, value in os.environ.items()})
    runtime['ncpu-logical'] = psutil.cpu_count(logical=True)
    runtime['ncpu-physical'] = psutil.cpu_count(logical=False)

    conf.run = OmegaConf.create(runtime)

    # add include paths
    if conf.opts.include:
        configuratt.PATH += list(conf.opts.include)
        log.info(f"added include paths: {' '.join(conf.opts.include)}")

    global CONFIG_DEPS
    CONFIG_DEPS = dependencies

    # check for missing requirements
    missing = configuratt.check_requirements(conf, [], strict=True)
    for (loc, name, _) in missing:
        log.warn(f"optional config section '{loc}' omitted due to unmet requirement '{name}'")

    return conf




from dataclasses import fields
from typing import List, Dict
import logging
import os.path
import time, datetime
import threading
from stimela.exceptions import BackendError
from stimela.utils.xrun_asyncio import dispatch_to_log
from stimela.task_stats import update_process_status
from stimela.kitchen.cab import Cab, Parameter

from kubernetes.client import ApiException
from kubernetes.stream import stream

from . import KubeBackendOptions, session_user_info, get_kube_api
from .kube_utils import apply_pod_spec
from stimela.backends.utils import resolve_remote_mounts

class PodProxy(object):

    def __init__(self, kube: KubeBackendOptions, podname: str, image_name: str, command: str, log: logging.Logger):
        self.namespace, self.kube_api, _ = get_kube_api()
        self.kube = kube
        self.name = podname
        self.log = log
        # setup user information -- current user info by default, overridden by backend options
        self.uinfo = KubeBackendOptions.UserInfo()
        for fld in fields(self.uinfo):
            value = getattr(kube.user, fld.name)
            setattr(self.uinfo, fld.name, value if value is not None else getattr(session_user_info, fld.name))

        # setup command
        if self.uinfo.inject_nss:
            command = \
                "cp /etc/passwd $HOME/.passwd; " + \
                "cp /etc/group $HOME/.group; " + \
                "echo $USER:x:$USER_UID:$USER_GID:$USER_GECOS:$HOME:/bin/sh >> $HOME/.passwd; " + \
                "echo $GROUP:x:$USER_GID: >> $HOME/.group; " + \
                command

        self.pod_spec = dict(
            containers = [dict(
                    image   = image_name,
                    imagePullPolicy = 'Always' if kube.always_pull_images else 'IfNotPresent',
                    name    = "job",
                    command = ["/bin/bash"],
                    args    = ["-c", command],
                    env     = [
                        dict(name="USER", value=self.uinfo.name),
                        dict(name="GROUP", value=self.uinfo.group),
                        dict(name="HOME", value=self.uinfo.home),
                        dict(name="USER_UID", value=str(self.uinfo.uid)),
                        dict(name="USER_GID", value=str(self.uinfo.gid)),
                        dict(name="USER_GECOS", value=str(self.uinfo.gecos))
                    ],
                    securityContext = dict(
                            runAsNonRoot = self.uinfo.uid!=0,
                            runAsUser = self.uinfo.uid,
                            runAsGroup = self.uinfo.gid,
                    ),
                    volumeMounts = [dict(
                        name = "home-directory",
                        mountPath = self.uinfo.home
                    )]
            )],
            volumes = [dict(
                name = "home-directory",
                emptyDir = dict(medium="Memory") if self.uinfo.home_ramdisk else {} 
            )],
            serviceAccountName = kube.service_account,
            automountServiceAccountToken = True,
            restartPolicy = "Never"
        )
        # apply predefined pod type specifications
        self.pod_spec = apply_pod_spec(kube.job_pod, self.pod_spec, kube.predefined_pod_specs, self.log, kind='job')

        # add runtime env settings
        for name, value in kube.env.items():
            value = os.path.expanduser(value)
            self.pod_spec['containers'][0]['env'].append(dict(name=name, value=value))

        self._session_init_container = None
        self._step_init_container = None          
        self._exit_logging_threads = False
        self._aux_pod_threads = {}
        self._mounts = {}
        # accumulate list of exceptions that are raised in the updater thread
        self._exceptions = []

    def _status_updater(self):
        while not self._exit_logging_threads:
            try:
                update_process_status()
            except BackendError as exc:
                self._exceptions.append(exc)
            time.sleep(1)

    def start_status_update_thread(self):
        if "status" not in self._aux_pod_threads:
            thread = threading.Thread(target=self._status_updater)
            self._aux_pod_threads["status"] = thread, None, "status update thread"
            thread.start()

    def check_status(self):
        """
        Checks if status thread has caught any exceptions, and re-reises them if so.
        Returns True of no exceptions.
        """
        if self._exceptions:
            if len(self._exceptions) == 1:
                raise self._exceptions[0]
            else:
                raise BackendError("k8s backend errors", *self._exceptions)
        return True

    @property
    def session_init_container(self):
        if self._session_init_container is None:
            self._session_init_container = dict(
                name="volume-session-init",
                image="quay.io/quay/busybox",
                command=["/bin/sh", "-c", ""],
                volumeMounts=[])
            self.pod_spec.setdefault('initContainers', []).append(self._session_init_container)
        return self._session_init_container

    @property
    def step_init_container(self):
        if self._step_init_container is None:
            self._step_init_container = dict(
                name="volume-step-init",
                image="quay.io/quay/busybox",
                command=["/bin/sh", "-c", ""],
                securityContext=dict(
                        runAsNonRoot = self.uinfo.uid!=0,
                        runAsUser = self.uinfo.uid,
                        runAsGroup = self.uinfo.gid),
                volumeMounts=[])
            self.pod_spec.setdefault('initContainers', []).append(self._step_init_container)
        return self._step_init_container
    
    @staticmethod
    def add_init_container_command(cont, command):
        cont['command'][-1] += command

    @staticmethod
    def add_init_container_mount(cont, volume_name, mount):
        for mnt in cont['volumeMounts']:
            if mnt['mountPath'] == mount:
                return
        cont['volumeMounts'].append(dict(
            name=volume_name, mountPath=mount
        ))

    def add_volume(self, volume_name, mount):
        if mount in self._mounts:
            raise BackendError(f"multiple volumes configured for mount {mount}")
        self.pod_spec['containers'][0]['volumeMounts'].append(dict(
            name=volume_name, mountPath=mount
        ))
        self._mounts[mount] = volume_name

    def add_volume_init(self, mount: str, commands: List[str], root: bool=False):
        """Adds volume initialization commands to init container"""
        # get approproate init container
        cont = self.session_init_container if root else self.step_init_container
        # add to its commands
        volume_name = self._mounts[mount]
        self.log.info(f"adding init commands for PVC '{volume_name}': {'; '.join(commands)}")
        self.add_init_container_command(cont, f"cd {mount}; {'; '.join(commands)}; ")
        self.add_init_container_mount(cont, volume_name, mount)

    def dispatch_container_logs(self, style: str, job: bool = True):
        containers = self.pod_spec.get('initContainers', [])
        if job:
            containers += self.pod_spec['containers']
        for cont in containers:
            contname = cont['name']
            try:
                loglines = self.kube_api.read_namespaced_pod_log(name=self.name, namespace=self.namespace, container=contname)
                for line in loglines.split("\n"):
                    if line:
                        dispatch_to_log(self.log, line, contname, "stdout", prefix=f"{contname}#", 
                                        style=style, output_wrangler=None)
            except ApiException as exc:
                dispatch_to_log(self.log, "no logs", contname, "stdout", prefix=f"{contname}#", 
                                style=style, output_wrangler=None)

    def _print_logs(self, name, style, container=None):
        """Helper function -- prints logs from a pod in a separate thread"""
        prefix = f"{name}#" if container is None else f"{name}.{container}#" 
        dispatch_to_log(self.log, f"started logging thread for {name} {container}", name, "stdout", prefix=prefix, style=style, output_wrangler=None)
        try:
            while not self._exit_logging_threads:
                try:
                    # rich.print(f"[yellow]started log thread for {name}[/yellow]")
                    # Open a stream to the logs
                    stream = self.kube_api.read_namespaced_pod_log(name=name, namespace=self.namespace, container=container, 
                                                                _preload_content=False, follow=True)
                    for line in stream.stream():
                        dispatch_to_log(self.log, line.decode().rstrip(), name, "stdout", prefix=prefix, style=style, output_wrangler=None)
                except ApiException as exc:
                    # dispatch_to_log(log, f"error reading log: {exc}", command_name, "stdout", prefix=prefix, style=style, output_wrangler=None)
                    pass
                time.sleep(2)
        finally:
            dispatch_to_log(self.log, f"exiting logging thread", name, "stdout", prefix=prefix, style=style, output_wrangler=None)
    
    def add_file_check_commands(self, params: Dict[str, Parameter], cab: Cab):
        # add commands to check for files and make directories
        must_exist_list, mkdir_list, remove_if_exists_list, active_mounts = \
            resolve_remote_mounts(params, cab.inputs, cab.outputs, cwd=self.kube.dir, mounts=set(self._mounts.keys()))
        
        if must_exist_list or mkdir_list or remove_if_exists_list:
            cont = self.step_init_container
            for path in mkdir_list:
                error = f"VALIDATION ERROR: mkdir {path} failed"
                self.add_init_container_command(cont, f"ls {path} >/dev/null; if mkdir -p {path}; then echo Created directory {path}; else echo {error}; exit 1; fi; ")
            for path in must_exist_list:
                error = f"VALIDATION ERROR: {path} doesn\\'t exist"
                self.add_init_container_command(cont, f"ls {path} >/dev/null; if test -e '{path}'; then echo Checking {path}: exists; else echo {error}; exit 1; fi; ")
            for path in remove_if_exists_list:
                self.add_init_container_command(cont, f"ls {path} >/dev/null; if test -e {path}; then echo Removing {path}; rm -fr {path}; true; fi; ")
            # make sure the relevant mounts are provided inside init container
            for mount in active_mounts:
                self.add_init_container_mount(cont, self._mounts[mount], mount)

    def start_logging_thread(self, name, style, container=None):
        """Starts thread to pick up logs"""
        thread = threading.Thread(target=self._print_logs, 
                        kwargs=dict(name=name, container=container, style=style))
        self._aux_pod_threads[name] = thread, name, f"auxiliary logging thread for {name}"
        thread.start()

    def initiate_cleanup(self):
        self._exit_logging_threads = True
        
    def cleanup(self):
        cleanup_time = datetime.datetime.now()
        def cleanup_elapsed():
            return (datetime.datetime.now() - cleanup_time).total_seconds()
        while self._aux_pod_threads and cleanup_elapsed() < 5:
            update_process_status()
            for name, (thread, aux_pod, desc) in list(self._aux_pod_threads.items()):
                if thread.is_alive():
                    thread.join(0.01)
                if thread.is_alive() and aux_pod:
                    self.log.info(f"{desc} alive, trying to delete associated pod")
                    try:
                        self.kube_api.delete_namespaced_pod(name=aux_pod, namespace=self.namespace)
                    except ApiException as exc:
                        self.log.warning(f"deleting pod {aux_pod} failed: {exc}")
                        pass
                if not thread.is_alive():
                    del self._aux_pod_threads[name]
            if self._aux_pod_threads:
                # self.log.info(f"{len(self._aux_pod_threads)} auxiliary threads still alive, sleeping for a bit")
                time.sleep(0.5)
        if self._aux_pod_threads:
            self.log.info(f"{len(self._aux_pod_threads)} auxiliary threads still alive after {cleanup_elapsed():.1}s, giving up on the cleanup")

    def run_pod_command(self, command, cmdname, input=None, wrangler=None):
        if type(command) is str:
            command = ["/bin/sh", "-c", command]
        has_input = bool(input)

        resp = stream(self.kube_api.connect_get_namespaced_pod_exec, self.name, self.namespace,
                    command=command,
                    stderr=True, stdin=has_input,
                    stdout=True, tty=False,
                    _preload_content=False)

        while resp.is_open():
            update_process_status()
            resp.update(timeout=1)
            if resp.peek_stdout():
                for line in resp.read_stdout().rstrip().split("\n"):
                    dispatch_to_log(self.log, line, cmdname, "stdout", output_wrangler=wrangler)
            if resp.peek_stderr():
                for line in resp.read_stderr().rstrip().split("\n"):
                    dispatch_to_log(self.log, line, cmdname, "stderr", output_wrangler=wrangler)
            if has_input:
                if input:
                    resp.write_stdin(input)
                    input = None
                else:
                    break

        retcode = resp.returncode
        resp.close()
        return retcode

            # # inject files into pod
            # if kube.inject_files:
            #     with declare_subcommand("configuring pod (inject)"):
            #         for filename, injection in kube.inject_files.items():
            #             content = injection.content
            #             formatter = InjectedFileFormatters.get(injection.format.name)
            #             if formatter is None:
            #                 raise StimelaCabParameterError(f"unsupported format {injection.format.name} for {filename}")
            #             # convert content to something serializable
            #             if isinstance(content, (DictConfig, ListConfig)):
            #                 content = OmegaConf.to_container(content)
            #             content = formatter(content)
            #             log.info(f"injecting {filename} into pod {podname}")
            #             retcode = run_pod_command(f"mkdir -p {os.path.dirname(filename)}; cat >{filename}", "inject", input=content)
            #             if retcode:
            #                 log.warning(f"injection returns exit code {retcode} after {elapsed()}")
            #             else:
            #                 log.info(f"injection successful after {elapsed()}")


            # if kube.pre_commands:
            #     with declare_subcommand("configuring pod (pre-commands)"):
            #         for pre_command in kube.pre_commands:
            #             log.info(f"running pre-command '{pre_command}' in pod {podname}")
            #             # calling exec and waiting for response
            #             retcode = run_pod_command(pre_command, pre_command.split()[0])
            #             if retcode:
            #                 log.warning(f"pre-command returns exit code {retcode} after {elapsed()}")
            #             else:
            #                 log.info(f"pre-command successful after {elapsed()}")



            # # add local mounts
            # def add_local_mount(name, path, dest, readonly):
            #     name = name.replace("_", "-")  # sanitize name
            #     pod_manifest['spec']['volumes'].append(dict(
            #         name = name,
            #         hostPath = dict(path=path, type='Directory' if os.path.isdir(path) else 'File')
            #     ))
            #     pod_manifest['spec']['containers'][0]['volumeMounts'].append(dict(name=name, mountPath=dest, readOnly=readonly))

            # # this will accumulate mounted paths from runtime spec
            # prior_mounts = {}

            # # add local mounts from runtime spec
            # for name, mount in kube.local_mounts.items():
            #     path = os.path.abspath(os.path.expanduser(mount.path))
            #     dest = os.path.abspath(os.path.expanduser(mount.dest)) if mount.dest else path
            #     if not os.path.exists(path) and mount.mkdir:
            #         pathlib.Path(path).mkdir(parents=True)
            #     add_local_mount(name, path, dest, mount.readonly)
            #     if path == dest:
            #         prior_mounts[path] = not mount.readonly

            # # add local mounts to support parameters
            # req_mounts = {} # resolve_required_mounts(params, cab.inputs, cab.outputs, prior_mounts=prior_mounts)
            # for i, (path, readwrite) in enumerate(req_mounts.items()):
            #     log.info(f"adding local mount {path} (readwrite={readwrite})")
            #     add_local_mount(f"automount-{i}", path, path, not readwrite)




from stimela import logger, LOG_FILE, BASE,  utils
from stimela.backends import docker, singularity, podman


def make_parser(subparsers):

    parser = subparsers.add_parser('pull', help='Pull docker stimela base images')

    add = parser.add_argument

    add("-im", "--image", nargs="+", metavar="IMAGE[:TAG]",
        help="Pull base image along with its tag (or version). Can be called multiple times")

    add("-f", "--force", action="store_true",
        help="force pull if image already exists")

    add("-s", "--singularity", action="store_true",
        help="Pull base images using singularity."
        "Images will be pulled into the directory specified by the enviroment varaible, STIMELA_PULLFOLDER. $PWD by default")

    add("-d", "--docker", action="store_true",
        help="Pull base images using docker.")

    add("-p", "--podman", action="store_true",
        help="Pull base images using podman.")

    add("-cb", "--cab-base", nargs="+",
        help="Pull base image for specified cab")

    add("-pf", "--pull-folder",
        help="Images will be placed in this folder. Else, if the environmnental variable 'STIMELA_PULLFOLDER' is set, then images will be placed there. "
        "Else, images will be placed in the current directory")

    parser.set_defaults(func=pull)


def pull(args, conf):
    pass



import re
from typing import Optional, Any, Union, Dict
import logging
from dataclasses import dataclass
from dataclasses import dataclass
from omegaconf import MISSING, OmegaConf, DictConfig
from omegaconf.errors import OmegaConfBaseException
from stimela.exceptions import CabValidationError
from scabha.exceptions import ScabhaBaseException
import stimela


class _BaseFlavour(object):
    """
    A flavour class represents a particular kind of runnable task
    (binary, python callable, inline python code, etc.)
    """
    def finalize(self, cab: "stimela.kitchen.cab.Cab"):
        """Finalizes flavour definition, given a cab"""
        self.command_name = cab.command.split()[0]

    def get_arguments(self, cab: "stimela.kitchen.cab.Cab", 
                            params: Dict[str, Any], 
                            subst: Dict[str, Any],
                            virtual_env: Optional[str] = None,
                            check_executable: bool = True,
                            log: Optional[logging.Logger] = None):
        """Returns command line arguments for running this flavour of task, given
        a cab and a set of parameters. 

        Args:
            cab (Cab):               cab definition
            params (Dict[str, Any]): parameter dict
            subst (Dict[str, Any]):  substitution namespace 
            virtual_env (Optional[str]): virtual environment to run in, or None
            check_executable (bool):  if True, cab may check for the executable to exist (but doesn't have to)
            log (Optional[Logger]):  optional logger


        Returns:
            Tuple[List, List]:       tuple of full_argument_list, abbreviated_argument_list
                                     The full list is meant for execution, the abbreviated list is meant for
                                     display and logging
        """
        pass


@dataclass
class _CallableFlavour(_BaseFlavour):
    """
    Represents a callable function (in python)
    """
    # if True, function returns dict of outputs
    output_dict: bool = False
    # if not None, function returns the named output
    output: Optional[str] = None

    @property
    def dict_outputs(self):
        return self.outputs == "{}"

    def finalize(self, cab: "stimela.kitchen.cab.Cab"):
        if self.output_dict and self.output:
            raise CabValidationError(f"cab {cab.name}: can't specify both 'output_dict' and 'output'")
        if self.output and self.output not in cab.outputs: 
            raise CabValidationError(f"cab {cab.name}: flavour.outputs='{self.outputs}' is not a known output")

_flavour_map = None
_flavour_schemas = None

def lookup_flavour(kind):
    """
    Given a "kind" string, looks up appropriate flavour class, and corresponding OmegeCaonf schema
    Returns:
        (class, OmegaConf.DictConfig): class and schema 
    """
    global _flavour_map, _flavour_schemas
    # init map the first time we're called
    if _flavour_map is None:
        from .binary import BinaryFlavour
        from .python_flavours import PythonCallableFlavour, PythonCodeFlavour
        from .casa import CasaTaskFlavour

        _flavour_map = {
            'binary': BinaryFlavour,
            'python': PythonCallableFlavour,
            'python-code': PythonCodeFlavour,
            'casa-task': CasaTaskFlavour
        }

        _flavour_schemas = {name: OmegaConf.structured(cls) for name, cls in _flavour_map.items()}

    if kind in _flavour_map:
        return _flavour_map[kind], _flavour_schemas[kind]
    else:
        return None, None

def init_cab_flavour(cab: "stimela.kitchen.cab.Cab"):
    """
    Given a cab definition, creates an object of the appropriate flavour class. 
    Cab.flavour can be specified as a string (default class definition), or as a mapping
    with a "kind" attribute, in which case flavour parameters may be passed.
    """
    flavour = cab.flavour
    if flavour is None:
        from .binary import BinaryFlavour
        flavour = BinaryFlavour()
    elif isinstance(flavour, str):
        cls, _ = lookup_flavour(flavour)
        if cls is None:
            raise CabValidationError(f"unknown flavour '{flavour}'")
        flavour = cls()
    elif isinstance(flavour, DictConfig):
        if 'kind' not in flavour:
            raise CabValidationError(f"flavour.kind not specified")
        cls, schema = lookup_flavour(flavour.kind)
        if cab is None:
            raise CabValidationError(f"unknown flavour.kind '{flavour.kind}'")
        try:
            defs = OmegaConf.merge(schema, flavour)
            flavour = cls(**defs)
        except (OmegaConfBaseException, ScabhaBaseException) as exc:
            raise CabValidationError(f"error in flavour definition", exc)
    else:
        raise CabValidationError(f"flavour must be a string or a mapping")
    flavour.finalize(cab)
    return flavour
